%{

#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <curses.h>
#include <ctype.h>
#include "common/memory.h"
#include "common/error.h"
#include "common/strbuf.h"
#include "common/hmap.h"
#include "common/intparse.h"
#include "common/llist.h"
#include "connmon.h"

static int color_index;
static int pair_index;
static int lineno;
static FILE *in;
static HMap *color_map;
static HMap *pair_map;
static LList *string_list;

static void yyerror(char *msg);
static int yylex(void);
static int lookup_pair(char *color1, char *color2);
static void create_color(char *name, short red, short green, short blue);

static const char *const style_names[] = {
	"normal", "standout", "underline", "reverse", "blink", "dim",
	"bold", "protect", "invisible", "altcharset",
};
static const int style_values[] = {
	A_NORMAL, A_STANDOUT, A_UNDERLINE, A_REVERSE, A_BLINK, A_DIM,
	A_BOLD, A_PROTECT, A_INVIS, A_ALTCHARSET, A_CHARTEXT,
};

%}

%token COLOR STRING INT FIELD STYLE
%union {
	int intval;
	char *strval;
}
%type <intval> COLOR INT FIELD STYLE styles style
%type <strval> STRING

%%

input:			color_decl
				| color_specs
				| /* empty */
				;
color_decl:		COLOR STRING INT INT INT	{ create_color($2,$3,$4,$5); }
				;
color_specs:	color_specs color_spec
				| color_spec
				;
color_spec:		FIELD styles				{ field_attrs[$1] = $2; }
				;
styles:			styles style				{ $$ = $1 | $2; }
				| /* empty */				{ $$ = 0; }
				;
style:			STRING STRING				{ $$ = lookup_pair($1, $2); }
				| STYLE						{ $$ = $1; }
				;

%%

static void
create_color(char *name, short red, short green, short blue)
{
	if (hmap_contains_key(color_map, name))
		die("%s: color already exists", name);

	if (color_index >= COLORS)
		die("%s: too many colors", name);

	init_color(color_index, red, green, blue);
	hmap_put(color_map, name, (void *) color_index);
	color_index++;
}

static int
lookup_pair(char *color1, char *color2)
{
	uint32_t c1;
	uint32_t c2;
	uint32_t index;

	if (!hmap_contains_key(color_map, color1))
		die("%s: no such color", color1);
	if (!hmap_contains_key(color_map, color2))
		die("%s: no such color", color2);

	c1 = (uint32_t) hmap_get(color_map, color1);
	c2 = (uint32_t) hmap_get(color_map, color2);

	index = (uint32_t) hmap_get(pair_map, (void *) (c1<<16|c2));
	if (index != 0)
		return COLOR_PAIR(index);

	if (pair_index >= COLOR_PAIRS-1)
		die("too many color pairs (max %d)", COLOR_PAIRS);
	index = pair_index++;
	init_pair(index, c1, c2);
	hmap_put(pair_map, (void *) (c1<<16|c2), (void *) index);

	return COLOR_PAIR(index);
}

static int
fgetch(FILE *in)
{
	int ch = fgetc(in);
	if (ch == EOF && ferror(in))
		die_errno("cannot read");
	if (ch == '\n')
		lineno++;
	return ch;
}

static int
yylex(void)
{
	int ch;
	int c;
	uint16_t value;
	char *str;

	while (TRUE) {
		ch = fgetch(in);
		if (feof(in))
			return EOF;
		if (ch == '#') {
			while (fgetch(in) != '\n' && !feof(in));
		} else if (!isspace(ch)) {
			break;
		}
	}

	str = strbuf_new();
	do {
		strbuf_append_char(&str, ch);
	} while ((ch = fgetch(in)) != EOF && !isspace(ch));

	if (parse_uint16(str, &value)) {
		yylval.intval = value;
		return INT;
	}

	if (strcmp(str, "color") == 0)
		return COLOR;

	for (c = 0; c < FIELD_COUNT; c++) {
		if (strcmp(str, field_names[c]) == 0) {
			yylval.intval = c;
			return FIELD;
		}
	}

	for (c = 0; c < sizeof(style_names)/sizeof(char *); c++) {
		if (strcmp(str, style_names[c]) == 0) {
			yylval.intval = style_values[c];
			return STYLE;
		}
	}

	yylval.strval = strbuf_free_to_string(&str);
	llist_add(string_list, yylval.strval);

	return STRING;
}

static void
yyerror(char *msg)
{
	die("%s in line %d", msg, lineno);
}

static uint32_t
hash_int(uint32_t value)
{
	return value;
}

static int32_t
compare_ints(uint32_t val1, uint32_t val2)
{
	return val1-val2;
}

bool
parse_color_scheme(const char *filename)
{
	int rc;

	in = fopen(filename, "r");
	if (in == NULL)
		die_errno("%s: cannot open file", filename);
	string_list = llist_new();
	lineno = 1;

	color_index = 0;
	color_map = hmap_new();
	hmap_put(color_map, "black", (void *) COLOR_BLACK);
	hmap_put(color_map, "red", (void *) COLOR_RED);
	hmap_put(color_map, "green", (void *) COLOR_GREEN);
	hmap_put(color_map, "yellow", (void *) COLOR_YELLOW);
	hmap_put(color_map, "blue", (void *) COLOR_BLUE);
	hmap_put(color_map, "magenta", (void *) COLOR_MAGENTA);
	hmap_put(color_map, "cyan", (void *) COLOR_CYAN);
	hmap_put(color_map, "white", (void *) COLOR_WHITE);

	pair_index = 1;
	pair_map = hmap_new();
	hmap_set_hash_function(pair_map, (HashFunc) hash_int);
	hmap_set_compare_function(pair_map, (CompareFunc) compare_ints);


	rc = yyparse();

	llist_iterate(string_list, free);
	hmap_free(color_map);
	hmap_free(pair_map);
	fclose(in);

	return (rc == 0);
}
