#include <config.h>
#include <curses.h>
#include <getopt.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <libconnmon.h>
#include <libhnl.h>
#include <libpnl.h>
#include <limits.h>
#include <signal.h>
#include <errno.h>
#include <adns.h>
#include "common/intparse.h"
#include "common/string-utils.h"
#include "common/memory.h"
#include "common/error.h"
#include "common/llist.h"
#include "common/time64.h"
#include "connmon.h"

enum {
	VERSION_OPT = 1000,
	HELP_OPT,
};

static const char *short_opts = "e:d:t:h:p:c:f:HPC";
static const struct option long_opts[] = {
	{ "driver",				required_argument,	NULL,	'e'			},
	{ "poll-delay",			required_argument,	NULL,	'd'			},
	{ "inactivity-time",	required_argument,	NULL,	't'			},
	{ "host-lookup",		required_argument,	NULL,	'h'			},
	{ "port-lookup",		required_argument,	NULL,	'p'			},
	{ "color-scheme",		required_argument,	NULL,	'c'			},
	{ "filter",				required_argument,	NULL,	'f'			},
	{ "no-host-lookup",		no_argument,		NULL,	'H'			},
	{ "no-port-lookup",		no_argument,		NULL,	'P'			},
	{ "no-caching",			no_argument,		NULL,	'C'			},
	{ "help", 				no_argument,		NULL,	HELP_OPT	},
	{ "version",			no_argument,		NULL,	VERSION_OPT	},
	{ 0, 					0, 					0,		0 			},
};

ConnMon *cm;
ConnList *oc_list;
ConnList *cc_list;
uint8_t inactivity_count = DEFAULT_INACTIVITY_COUNT;

static HNLManager *hnl;
static PNLManager *pnl;

static void
connection_created(ConnTrack *ct)
{
	ConnTrackData *data = xmalloc(sizeof(ConnTrackData));

	ct->data = data;
	data->ct = ct;
	data->pos = -1;
	data->state = CTD_STATE_CREATED;
	data->rotator = 99;			/* FIXME: why 99 (=UINT8_MAX-1?) and not 0? */
	data->inactivity = UINT8_MAX;
	data->src_name = NULL;
	data->dst_name = NULL;
	data->src_port = pnl_lookup(pnl, ct->tuple.protocol, ct->tuple.ps.tcp_udp.src_port);
	data->dst_port = pnl_lookup(pnl, ct->tuple.protocol, ct->tuple.ps.tcp_udp.dst_port);
	data->src_looked = FALSE;
	data->dst_looked = FALSE;
	data->print_time = -1;

	data->src_query = hnl_lookup(hnl, ct->tuple.src_ip, data);
	if (ct->tuple.src_ip != ct->tuple.dst_ip) {
		data->dst_query = hnl_lookup(hnl, ct->tuple.dst_ip, data);
	} else {
		data->dst_query = data->src_query;
	}
}

static void
lookup_finished(HNLManager *hm, uint32_t ip, char *result, void *userdata)
{
	ConnTrackData *data = userdata;

	if (ip == data->ct->tuple.src_ip) {
		data->src_name = (result == NULL ? NULL : xstrdup(result));
		data->src_query = NULL;
		data->src_looked = TRUE;
	}
	if (ip == data->ct->tuple.dst_ip) {
		data->dst_name = (result == NULL ? NULL : xstrdup(result));
		data->dst_query = NULL;
		data->dst_looked = TRUE;
	}

	connmon_ui_redisplay_connection(data);
}

static bool
is_connection_open(ConnTrack *ct)
{
	return ct->tuple.protocol != CM_PROTO_TCP
			|| ct->tcp_state == CM_TCP_STATE_ESTABLISHED
			|| ct->tcp_state == CM_TCP_STATE_SYN_SENT
			|| ct->tcp_state == CM_TCP_STATE_SYN_RECV
			|| ct->tcp_state == CM_TCP_STATE_LISTEN;
}

static void
connection_updated(ConnTrack *ct)
{
	ConnTrackData *data = ct->data;

	if (is_connection_open(ct)) {
		if (data->state != CTD_STATE_OPEN) {
			data->state = CTD_STATE_OPEN;
			append_connection(oc_list, data);
		}
	} else {
		if (data->state != CTD_STATE_CLOSED) {
			data->state = CTD_STATE_CLOSED;
			append_connection(cc_list, data);
		}
	}
}

static void
connection_deleted(ConnTrack *ct)
{
	ConnTrackData *data = ct->data;
	data->state = CTD_STATE_REMOVED;
}

void
free_connection(ConnTrackData *data)
{
	if (data->dst_query != NULL && data->src_query != data->dst_query) {
		hnl_cancel(hnl, data->dst_query);
		data->dst_query = NULL;
	}
	if (data->src_query != NULL) {
		hnl_cancel(hnl, data->src_query);
		data->src_query = NULL;
	}
	if (data->src_name != NULL)
		free(data->src_name);
	if (data->dst_name != NULL)
		free(data->dst_name);
	free(data);
}

static void
sigwinch_handler(int sig)
{
	connmon_ui_resized();
}

static void
display_version(void)
{
	printf("%s %s\n\
Written by %s.\n\
\n\
Copyright (C) %s.\n\
This is free software; see the source for copying conditions.  There is NO\n\
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
		PACKAGE, VERSION, AUTHORS, COPYRIGHT);
}

static void
display_help(void)
{
	printf("Usage: %s [OPTION]...\n\
Start ConnMon, a connection and bandwidth monitoring program with a\n\
console-based used interface.\n\
\n\
  -c, --color-scheme=FILE      specify a different color scheme\n\
  -f, --filter=EXPR            only show connections matching filter\n\
  -e, --driver=SPEC            connection list driver\n\
  -d, --poll-delay=TIME        time between polls of connections\n\
  -t, --inactivity-time=TIME   time after which a connection is inactive\n\
  -h, --host-lookup=SPEC       a comma-separated list of lookup services\n\
  -p, --port-lookup=SPEC       a colon-separated list of port name files\n\
  -H, --no-host-lookup         do not look up host names\n\
  -P, --no-port-lookup         do not look up port names\n\
  -C, --no-caching             do not cache ip to host name mappings\n\
      --help                   display this help and exit\n\
      --version                output version information and exit\n\
\n\
The default poll time is 300ms, and default inactivity time is 1200ms.\n\
TIME may be specified with a unit, for example 1min. Possible units are\n\
s (microseconds), ms (milliseconds), s or sec (seconds), or min (minutes).\n\
The default unit (if none specified) is s.\n\
\n\
The host lookup service specification is in the format `SERVICE[:ARGS][,..]'.\n\
The default specification is `hosts,adns'. Possible services are `hosts'\n\
(/etc/hosts-like lookup), `adns' (dns lookup with GNU adns) and `ghba'\n\
(experminal lookup with gethostbyaddr). The `hosts' service takes an optional\n\
argument specifying the file containing ip to host name-mappings (`/etc/hosts'\n\
is assumed if none specified). Caching must be enabled with the ghba service\n\
or the system will be overflooded by connections.\n\
\n\
The default port name file is `/etc/services'.\n\
\n\
The default driver is `netfilter'. A driver specification is in the format\n\
`DRIVER[:ARGS]'. Possible drivers are `netfilter' and `linux'.\n\
\n\
Report bugs to <%s>.\n", program_invocation_name, BUG_EMAIL);
}

static bool
parse_time(const char *timespec, uint64_t *time)
{
	uint64_t multiple = 1;
	int len = strlen(timespec);
	char buf[len+1];

	strcpy(buf, timespec);
	if (ends_with(timespec, "s")) {
		buf[len-2] = '\0';
		multiple = 1;
	} else if (ends_with(timespec, "ms")) {
		buf[len-2] = '\0';
		multiple = 1000;
	} else if (ends_with(timespec, "s")) {
		buf[len-1] = '\0';
		multiple = 1000*1000;
	} else if (ends_with(timespec, "sec")) {
		buf[len-3] = '\0';
		multiple = 1000*1000;
	} else if (ends_with(timespec, "min")) {
		buf[len-3] = '\0';
		multiple = 1000*1000*60;
	}

	if (!parse_uint64(buf, time))
		return FALSE;

	*time *= multiple;
	return TRUE;
}

static void
end_ncurses(void)
{
	endwin();
}

int
main(int argc, char **argv)
{
	uint64_t inactivity_time = 0;
	uint64_t sleep_time = DEFAULT_SLEEP_TIME;
	int64_t wakeup_time;
	char *hnl_services = DEFAULT_HNL_SERVICES;
	char *pnl_files = DEFAULT_PNL_FILES;
	bool hnl_caching = TRUE;
	char *color_scheme = NULL;
	char *filter = NULL;
	char *driver = NULL;
	int c;

	while ((c = getopt_long(argc, argv, short_opts, long_opts, NULL)) != -1) {
		switch (c) {
		case 'f':
			filter = optarg;
			break;
		case 'e':
			driver = optarg;
			break;
		case 'd':
			if (!parse_time(optarg, &sleep_time))
				die("invalid time specification: %s", optarg);
			break;
		case 't':
			if (!parse_time(optarg, &inactivity_time))
				die("invalid time specification: %s", optarg);
			if (inactivity_time == 0)
				die("inactivity time out of range");
			break;
		case 'h':
			hnl_services = optarg;
			break;
		case 'H':
			hnl_services = "";
			break;
		case 'p':
			pnl_files = optarg;
			break;
		case 'P':
			pnl_files = "";
			break;
		case 'C':
			hnl_caching = FALSE;
			break;
		case 'c':
			color_scheme = optarg;
			break;
		case HELP_OPT:
			display_help();
			exit(0);
		case VERSION_OPT:
			display_version();
			exit(0);
		case '?':
			exit(1);
		}
	}
	if (inactivity_time != 0) {
		inactivity_time /= sleep_time;
		if (inactivity_time == 0 || inactivity_time > UINT8_MAX)
			die("inactivity time out of range");
		inactivity_count = inactivity_time;
	}

	hnl = hnl_new();
	hnl->lookup_finished = lookup_finished;
	hnl_set_caching(hnl, hnl_caching);
	if (!hnl_initialize_from_spec(hnl, hnl_services))
		die_error();

	pnl = pnl_new();
	if (!pnl_initialize_from_spec(pnl, pnl_files))
		die_error();

	cm = cm_new(driver);
	if (cm == NULL)
		die_error();
	if (filter != NULL)
		cm_set_filter(cm, filter);
	cm->create_handler = connection_created;
	cm->delete_handler = connection_deleted;
	cm->update_handler = connection_updated;
	oc_list = create_connection_list(CTD_STATE_OPEN, "Open Connections");
	cc_list = create_connection_list(CTD_STATE_CLOSED, "Closed Connections");

	program_termination_hook = end_ncurses;
	connmon_ui_initialize();

	/* We need to register our own SIGWINCH handler because
	 * the default one provided by ncurses doesn't work
	 * with the select() event model below.
	 * This is because
	 *  1) ungetch(KEY_RESIZE) is done when SIGWINCH is catched.
	 *     It doesn't do any IO, and select isn't (for reasons
	 *     unknown) interrupted.
	 */
	signal(SIGWINCH, sigwinch_handler);

	if (color_scheme != NULL && !parse_color_scheme(color_scheme))
		exit(1);

	if (!cm_read_conns(cm))
		die_error();
	wakeup_time = time64_now() + sleep_time;
	connmon_ui_redraw();

	while (TRUE) {
		SelectData sd;
		int64_t now_time;
		int rc;

		hnl_clear_select_data(&sd);
		FD_SET(STDIN_FILENO, &sd.read_fds);
		sd.max_fd = STDIN_FILENO + 1;

		now_time = time64_now();
		time64_to_timeval(&sd.timeout, max(0, wakeup_time-now_time));
		hnl_preselect(hnl, &sd);
		rc = select(sd.max_fd, &sd.read_fds, &sd.write_fds, &sd.except_fds, &sd.timeout);
		if (rc < 0 && errno != EINTR)
			die(NULL);

		now_time = time64_now();
		hnl_postselect(hnl, &sd);

		if (FD_ISSET(STDIN_FILENO, &sd.read_fds)) {
			if (!connmon_ui_read_input())
				break;
		}
		if (now_time >= wakeup_time) {
			if (!cm_read_conns(cm))
				die_error();
			wakeup_time = now_time + sleep_time;
			connmon_ui_connections_updated();
		}

		connmon_ui_redraw();
	}

	connmon_ui_finish();
	clean_connection_list(oc_list);	/* FIXME: not necessary */
	clean_connection_list(cc_list);	/* FIXME: not necessary */
	free_connection_list(oc_list);
	free_connection_list(cc_list);
	hnl_free(hnl);
	cm_free(cm);
	exit(0);
}
