#include <limits.h>
#include <unistd.h>
#include <curses.h>
#include <termios.h>
#include <sys/ioctl.h>
#include "curses-util.h"
#include "connmon.h"

const char *const field_names[FIELD_COUNT] = {
	"main_header_active", "main_header_inactive", "main_header_total",
	"main_header_top", "main_footer_slash", "main_footer_key",
	"main_footer_colon", "main_footer_desc", "main_background",
	"help_header", "help_footer", "help_background",
	"conn_host_ip", "conn_host_name", "conn_host_dots",
	"conn_host_pending_ip", "conn_host_pending_parens", "conn_proto_tcp",
	"conn_proto_udp", "conn_proto_icmp", "conn_proto_ip",
	"conn_port_number", "conn_port_name", "conn_port_colon",
	"conn_direction", "conn_rate_value", "conn_rate_unavailable",
	"conn_rate_unit", "conn_rotator", "conn_tcp_state",
	"conn_icmp_type", "connlist_background", "total_conn_count",
	"total_conn_text", "total_background", "main_footer_background",
};
int field_attrs[FIELD_COUNT] = {
	COLOR_PAIR(2) | A_BOLD, COLOR_PAIR(1), COLOR_PAIR(1),
	COLOR_PAIR(4) | A_BOLD, 0, A_BOLD,
	0, COLOR_PAIR(4) | A_BOLD, COLOR_PAIR(1),
	COLOR_PAIR(2) | A_BOLD, COLOR_PAIR(2) | A_BOLD, COLOR_PAIR(5),
	0, 0, 0,
	0, 0, 0,
	0, 0, 0,
	0, 0, 0,
	0, 0, 0,
	0, 0, 0,
	0, COLOR_PAIR(5), 0,
	0, COLOR_PAIR(5), 0,
};

static ConnList *cur_list;
static WINDOW *total_window;
static bool redraw_main;
static bool redraw_total;
static bool displaying_help;

static char *help_keys[] = {
	"Scroll",			"Up", "Dn", "Home", "End", NULL,
	"Switch Pane",		"Tab", NULL,
	"Pane Up/Down",		"R", "F", NULL,
	"Help",				"H", NULL,
	"Quit",				"Q", NULL,
	NULL,
};

void
connmon_ui_connections_updated(void)
{
	oc_list->redraw_in = oc_list->redraw_out = TRUE;
	cc_list->redraw_in = cc_list->redraw_out = TRUE;
	redraw_total = TRUE;
}

static void
redraw_totals(ConnMon *cm, WINDOW *win)
{
	if (!redraw_total)
		return;

	redraw_total = FALSE;
	display_totals(cm, win);
}

void
connmon_ui_draw_header(int y, char *title, int count, int start, int height, bool iscur)
{
	char buf[40];
	int newpos;

	newpos = COLS - sprintf(buf, " %d..%d of %d ",
			min(count, start+1), start + min(count, height), count);

	attron(iscur ? ATTR(MAIN_HEADER_ACTIVE) : ATTR(MAIN_HEADER_INACTIVE));
	move(y, 0);
	addch(iscur ? ACS_BULLET : ' ');
	addch(' ');
	addnstr(title, min(strlen(title), COLS-3));
	if (newpos < 0) {
		mvaddnstr(y, 0, buf, COLS);
	} else if (newpos < getcurx(stdscr)) {
		mvaddstr(y, newpos, buf);
	} else {
		fillx(newpos);
		addstr(buf);
	}
	attroff(iscur ? ATTR(MAIN_HEADER_ACTIVE) : ATTR(MAIN_HEADER_INACTIVE));
}

static void
redraw_list_exterior(ConnList *cl)
{
	if (cl->redraw_out && cl->height != 0) {
		connmon_ui_draw_header(cl->y - 1, cl->title,
				llist_size(cl->connections), cl->start_pos, cl->height,
				cl == cur_list);
		cl->redraw_out = FALSE;
		refresh();
	}
}

void
connmon_ui_redraw_key_help(int y, char **keys)
{
	int c;

	attron(ATTR(MAIN_FOOTER_BACKGROUND));
	mvaddch(y, 0, ' ');
	attroff(ATTR(MAIN_FOOTER_BACKGROUND));
	for (c = 0; keys[c] != NULL; c++) {
		int d;

		if (c != 0) {
			attron(ATTR(MAIN_FOOTER_BACKGROUND));
			addstr("  ");
			attroff(ATTR(MAIN_FOOTER_BACKGROUND));
		}
		for (d = c+1; keys[d] != NULL; d++) {
			if (d != c+1) {
				attron(ATTR(MAIN_FOOTER_SLASH));
				addch('/');
				attroff(ATTR(MAIN_FOOTER_SLASH));
			}
			attron(ATTR(MAIN_FOOTER_KEY));
			addstr(keys[d]);
			attroff(ATTR(MAIN_FOOTER_KEY));
		}
		attron(ATTR(MAIN_FOOTER_COLON));
		addch(':');
		attroff(ATTR(MAIN_FOOTER_COLON));
		attron(ATTR(MAIN_FOOTER_DESC));
		addstr(keys[c]);
		attroff(ATTR(MAIN_FOOTER_DESC));
		
		c = d;
	}
	attron(ATTR(MAIN_FOOTER_BACKGROUND));
	fillx(COLS);
	attroff(ATTR(MAIN_FOOTER_BACKGROUND));
}

static void
redraw_main_window(void)
{
	if (!redraw_main)
		return;

	redraw_main = FALSE;
	attron(ATTR(MAIN_HEADER_TOP));
	mvaddch(0,0, ' ');
	printw(" ConnMon " VERSION " ");
	fillx(COLS);
	attroff(ATTR(MAIN_HEADER_TOP));

	connmon_ui_redraw_key_help(LINES-1, help_keys);

	attron(ATTR(MAIN_HEADER_TOTAL));
	move(LINES-3, 0);
	fillx(2);
	addstr("Total");
	fillx(COLS);
	attroff(ATTR(MAIN_HEADER_TOTAL));

	refresh();
}

static void
move_separator_up(void)
{
	if (cc_list->height == 0) {
		modify_connection_list_bounds(oc_list, -2, 0);
		modify_connection_list_bounds(cc_list, 1, -1);
		oc_list->redraw_in = oc_list->redraw_out = TRUE;
		cc_list->redraw_in = cc_list->redraw_out = TRUE;
	} else if (oc_list->height == 1) {
		modify_connection_list_bounds(oc_list, -1, 0);
		modify_connection_list_bounds(cc_list, 2, -2);
		cur_list = cc_list;
		oc_list->redraw_in = oc_list->redraw_out = TRUE;
		cc_list->redraw_in = cc_list->redraw_out = TRUE;
	} else if (oc_list->height > 1) {
		modify_connection_list_bounds(oc_list, -1, 0);
		modify_connection_list_bounds(cc_list, 1, -1);
		oc_list->redraw_in = oc_list->redraw_out = TRUE;
		cc_list->redraw_in = cc_list->redraw_out = TRUE;
	}
}

static void
move_separator_down(void)
{
	if (oc_list->height == 0) {
		modify_connection_list_bounds(oc_list, 1, 0);
		modify_connection_list_bounds(cc_list, -2, 2);
		oc_list->redraw_in = oc_list->redraw_out = TRUE;
		cc_list->redraw_in = cc_list->redraw_out = TRUE;
	} else if (cc_list->height == 1) {
		modify_connection_list_bounds(oc_list, 2, 0);
		modify_connection_list_bounds(cc_list, -1, 1);
		cur_list = oc_list;
		oc_list->redraw_in = oc_list->redraw_out = TRUE;
		cc_list->redraw_in = cc_list->redraw_out = TRUE;
	} else if (cc_list->height > 1) {
		modify_connection_list_bounds(oc_list, 1, 0);
		modify_connection_list_bounds(cc_list, -1, 1);
		oc_list->redraw_in = oc_list->redraw_out = TRUE;
		cc_list->redraw_in = cc_list->redraw_out = TRUE;
	}
}

static void
switch_panes(void)
{
	ConnList *other_list;

	other_list = (cur_list == oc_list ? cc_list : oc_list);
	if (other_list->height != 0) {
		cur_list = other_list;
		oc_list->redraw_out = TRUE;
		cc_list->redraw_out = TRUE;
	} else {
		beep();
	}
}

static void
set_default_window_sizes(void)
{
	oc_list->height = (LINES-6)/2 + (LINES-6)%2;
	oc_list->y = 2;
	cc_list->height = (LINES-6)/2;
	cc_list->y = oc_list->y + oc_list->height + 1;
}

static void
update_window_sizes(void)
{
	if (oc_list->height == 0) {
		cc_list->height = LINES-5;
	}
	else if (cc_list->height == 0) {
		oc_list->height = LINES-5;
		cc_list->y = LINES-3;
	}
	else {
		if (cc_list->y < LINES-3) {
			cc_list->height = LINES - cc_list->y - 3;
		} else if (LINES >= 8)  {
			cc_list->y = LINES-4;
			cc_list->height = 1;
			oc_list->height = LINES-7;
		} else if (cur_list == oc_list) {
			cc_list->height = 0;
			cc_list->y = LINES-3;
			oc_list->height = LINES-5;
		} else if (cur_list == cc_list) {
			oc_list->height = 0;
			cc_list->y = 2;
			cc_list->height = LINES-5;
		}
	}
}

static void
create_windows(void)
{
	if (oc_list->window != NULL)
		delwin(oc_list->window);

	update_window_sizes();

	oc_list->window = newwin(oc_list->height, 0, oc_list->y, 0);
	leaveok(oc_list->window, TRUE);
	wbkgdset(oc_list->window, ATTR(CONNLIST_BACKGROUND));
	werase(oc_list->window);
	oc_list->redraw_in = TRUE;
	oc_list->redraw_out = TRUE;

	if (cc_list->window != NULL)
		delwin(cc_list->window);

	cc_list->window = newwin(cc_list->height, 0, cc_list->y, 0);
	leaveok(cc_list->window, TRUE);
	wbkgdset(cc_list->window, ATTR(CONNLIST_BACKGROUND));
	werase(cc_list->window);
	cc_list->redraw_in = TRUE;
	cc_list->redraw_out = TRUE;

	if (total_window != NULL)
		delwin(total_window);

	total_window = newwin(1, 0, cc_list->y+cc_list->height+1, 0);
	leaveok(total_window, TRUE);
	wbkgdset(total_window, ATTR(TOTAL_BACKGROUND));
	werase(total_window);

	redraw_total = TRUE;
	redraw_main = TRUE;
}

void
connmon_ui_initialize(void)
{
	initscr();
	cbreak();
	noecho();
	nodelay(stdscr, TRUE);
	keypad(stdscr, TRUE);
	leaveok(stdscr, TRUE);
//	scrollok(stdscr, FALSE);
	start_color();
	init_pair(1, COLOR_WHITE, COLOR_BLUE);
	init_pair(2, COLOR_WHITE, COLOR_MAGENTA);
	init_pair(3, COLOR_WHITE, COLOR_CYAN);
	init_pair(4, COLOR_YELLOW, COLOR_BLUE);
	init_pair(5, COLOR_BLACK, COLOR_WHITE);
	init_pair(6, COLOR_GREEN, COLOR_WHITE);
	init_pair(7, COLOR_RED, COLOR_WHITE);
	init_pair(8, COLOR_BLUE, COLOR_WHITE);
	bkgdset(ATTR(MAIN_BACKGROUND));
	erase();
	curs_set(0);
	
	total_window = NULL;
	set_default_window_sizes();
	create_windows();
	cur_list = oc_list;
	displaying_help = FALSE;
}

void
connmon_ui_resized(void)
{
	struct winsize size;
	if (ioctl(STDIN_FILENO, TIOCGWINSZ, &size) == 0) {
		resizeterm(size.ws_row, size.ws_col);

		/* resizeterm always used to unget KEY_RESIZE, but in
		 * more recent versions of ncurses it only does that
		 * if WINCH was cault - by ncurses itself. Oh well,
		 * it doesn't matter that much if the screen is redrawn
		 * twice after resize.
		 */
		ungetch(KEY_RESIZE);
	}
}

void
connmon_ui_redraw(void)
{
	if (displaying_help) {
		connmon_ui_help_redraw();
	} else {
		redraw_main_window();
		redraw_totals(cm, total_window);
		redraw_connection_list(oc_list);
		redraw_connection_list(cc_list);
		redraw_list_exterior(oc_list);
		redraw_list_exterior(cc_list);
	}
}

bool
connmon_ui_read_input(void)
{
	int ch;

	if (displaying_help) {
		if (connmon_ui_help_read_input())
			return TRUE;

		displaying_help = FALSE;
		set_default_window_sizes();
		create_windows();
		return TRUE;
	}

	while ((ch = getch()) != ERR) {
		if (ch == KEY_UP) {
			scroll_connection_list(cur_list, -1);
		} else if (ch == KEY_DOWN) {
			scroll_connection_list(cur_list, 1);
		} else if (ch == KEY_HOME) {
			scroll_connection_list(cur_list, INT_MIN);
		} else if (ch == KEY_END) {
			scroll_connection_list(cur_list, INT_MAX);
		} else if (ch == KEY_NPAGE) {
			scroll_connection_list(cur_list, max(1, cur_list->height-1));
		} else if (ch == KEY_PPAGE) {
			scroll_connection_list(cur_list, -max(1, cur_list->height-1));
		} else if (ch == '\t') {
			switch_panes();
		} else if (ch == 'r') {
			move_separator_up();
		} else if (ch == 'f') {
			move_separator_down();
		} else if (ch == 'h') {
			displaying_help = TRUE;
			connmon_ui_help_display();
		} else if (ch == KEY_RESIZE) {
			create_windows();
		} else if (ch == 'q') {
			return FALSE;
		}
	}

	return TRUE;
}

void
connmon_ui_finish(void)
{
	endwin();
}

void
connmon_ui_redisplay_connection(ConnTrackData *data)
{
	if (displaying_help)
		return;

	if (data->state == CTD_STATE_OPEN)
		redisplay_connection_safe(oc_list, data);
	else if (data->state == CTD_STATE_CLOSED)
		redisplay_connection_safe(cc_list, data);
}
