#include <config.h>
#include <curses.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "common/time64.h"
#include "common/string-utils.h"
#include "curses-util.h"
#include "connmon.h"

#define ATTR_CONN_PROTO(x) \
	((x) == CM_PROTO_TCP  ? ATTR(CONN_PROTO_TCP)  : \
	((x) == CM_PROTO_UDP  ? ATTR(CONN_PROTO_UDP)  : \
	((x) == CM_PROTO_ICMP ? ATTR(CONN_PROTO_ICMP) : ATTR(CONN_PROTO_IP))))

#define HIPQUAD(addr) \
	((unsigned char *) &addr)[3], \
	((unsigned char *) &addr)[2], \
	((unsigned char *) &addr)[1], \
	((unsigned char *) &addr)[0]

static uint8_t total_rotator = 0;
static uint8_t total_inactivity = UINT8_MAX;
static int64_t total_print_time = -1;
static const char rotation[] = { '|', '/', '-', '\\', '|', '/', '-', '\\' };
static const char *icmp_types[] = {
	/*  0 -  3 */ "ECHOREPLY", NULL, NULL, "DST_UNREACH",
	/*  4 -  7 */ "SRC_QUENCH", "REDIRECT", NULL, NULL,	
	/*  8 - 11 */ "ECHO", NULL, NULL, "TIME_EXCEED",
	/* 12 - 15 */ "PARAM_PROB", "TIMESTAMP", "TS_REPLY", "INFO_REQ",
	/* 16 - 18 */ "INFO_REPLY", "ADDRESS", "ADDR_REPLY",
};

static char *
get_connection_state(ConnTrack *ct)
{
	if (ct->tuple.protocol != CM_PROTO_TCP)
		return "";

	switch (ct->tcp_state) {
	case CM_TCP_STATE_ESTABLISHED:
		return "ESTABLISHED";
	case CM_TCP_STATE_SYN_SENT:
		return "SYN_SENT";
	case CM_TCP_STATE_SYN_RECV:
		return "SYN_RECV";
	case CM_TCP_STATE_FIN_WAIT:
		return "FIN_WAIT";
	case CM_TCP_STATE_TIME_WAIT:
		return "TIME_WAIT";
	case CM_TCP_STATE_CLOSE:
		return "CLOSE";
	case CM_TCP_STATE_CLOSE_WAIT:
		return "CLOSE_WAIT";
	case CM_TCP_STATE_LAST_ACK:
		return "LAST_ACK";
	case CM_TCP_STATE_LISTEN:
		return "LISTEN";
	case CM_TCP_STATE_NONE:
	default:
		return "NONE";
	}
}

static char *
get_connection_type(uint16_t protocol)
{
	if (protocol == CM_PROTO_TCP)
		return "tcp";
	if (protocol == CM_PROTO_UDP)
		return "udp";
	if (protocol == CM_PROTO_ICMP)
		return "icmp";
	return "ip";
}

void
clear_line(WINDOW *window, int line)
{
	wmove(window, line, 0);
	wclrtoeol(window);
}

static int64_t
time_elapsed(int64_t time1, int64_t time2)
{
	return (time1 != 0 && time2 != 0 && time1 < time2 ? time2 - time1 : 0);
}

static char
get_rate_char(int64_t time, int64_t src_bytes, int64_t dst_bytes,
			  uint8_t *rotator, uint8_t *inactivity, bool updated)
{
	if (time != 0) {
		if (src_bytes != 0 || dst_bytes != 0) {
			if (updated) {
				*rotator = (*rotator+1) % sizeof(rotation);
				*inactivity = 0;
			}
			return rotation[*rotator];
		}
		if (*inactivity < inactivity_count) {
			if (updated)
				(*inactivity)++;
			return rotation[*rotator];
		}
	}

	return '.';
}

static int written;

static void
safe_addnstr(WINDOW *win, int attr, const char *str, int n)
{
	if (written < COLS) {
		wattron(win, attr);
		waddnstr(win, str, min(n, COLS - written));
		wattroff(win, attr);
		written += strnlen(str, min(n, COLS - written));
	}
}

static void
safe_addstr(WINDOW *win, int attr, const char *str)
{
	if (written < COLS) {
		wattron(win, attr);
		waddnstr(win, str, COLS - written);
		wattroff(win, attr);
		written += strnlen(str, COLS-written);
	}
}

static void
safe_printw(WINDOW *win, int attr, char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (written < COLS) {
		wattron(win, attr);
		written += vwnprintw(win, COLS - written, fmt, ap);
		wattroff(win, attr);
	}
	va_end(ap);
}

static void
safe_fillx(WINDOW *win, int x)
{
	if (written < x && written < COLS) {
		wfillx(win, x);
		written = x;
	}
}

static void
put_host(WINDOW *win, char *name, bool looked, uint32_t ip, int protocol,
		 uint16_t port, int maxlen, const char *portname)
{
	int portlen = 0;

	if (protocol == CM_PROTO_TCP || protocol == CM_PROTO_UDP)
		portlen = (portname == NULL ? 1+uintlen(port) : 1+strlen(portname));

	/* We assume below that we have space for at least '(..):'. */
	/* If the port name is too long, we must shorten it. */
	if (portlen+5 >= maxlen)
		portlen -= 5;

	/* We display a host in three different ways: */
	/* - with hostname */
	/* - with ip (knowning that there is no corresponding hostname) */
	/* - with ip (knowning that the hostname has not been looked up yet) */
	if (name != NULL) {
		if (strlen(name) + portlen > maxlen) {
			safe_addnstr(win, ATTR(CONN_HOST_NAME), name, maxlen-portlen-2);
			safe_addstr(win, ATTR(CONN_HOST_DOTS), "..");
		} else {
			safe_addstr(win, ATTR(CONN_HOST_NAME), name);
		}
	} else if (looked) {
		char namebuf[16];
		sprintf(namebuf, "%u.%u.%u.%u", HIPQUAD(ip));
		if (strlen(namebuf) + portlen > maxlen) {
			safe_addnstr(win, ATTR(CONN_HOST_IP), namebuf, maxlen-portlen-2);
			safe_addstr(win, ATTR(CONN_HOST_DOTS), "..");
		} else {
			safe_addstr(win, ATTR(CONN_HOST_IP), namebuf);
		}
	} else {
		char namebuf[16];
		sprintf(namebuf, "%u.%u.%u.%u", HIPQUAD(ip));
		safe_addstr(win, ATTR(CONN_HOST_PENDING_PARENS), "(");
		if (strlen(namebuf) + 2 + portlen > maxlen) {
			safe_addnstr(win, ATTR(CONN_HOST_PENDING_IP), namebuf, maxlen-portlen-4);
			safe_addstr(win, ATTR(CONN_HOST_DOTS), "..");			/* FIXME */
		} else {
			safe_addstr(win, ATTR(CONN_HOST_PENDING_IP), namebuf);
		}
		safe_addstr(win, ATTR(CONN_HOST_PENDING_PARENS), ")");
	}

	if (protocol == CM_PROTO_TCP || protocol == CM_PROTO_UDP) {
		safe_addstr(win, ATTR(CONN_PORT_COLON), ":");
		if (portname != NULL) {
			safe_addnstr(win, ATTR(CONN_PORT_NAME), portname, portlen);
		} else {
			safe_printw(win, ATTR(CONN_PORT_NUMBER), "%u", port);
		}
	}
}

static void
put_single_rate_spec(WINDOW *win, int64_t bytes, int64_t time)
{
	if (time == 0) {
		safe_addstr(win, ATTR(CONN_RATE_UNAVAILABLE), "  --");
		safe_addstr(win, 0, " ");
		safe_addstr(win, ATTR(CONN_RATE_UNIT), "b/s");
	} else {
		int64_t rate = bytes * MICROS_PER_SEC/time;
		if (rate <= 1024) {
			safe_printw(win, ATTR(CONN_RATE_VALUE), "%4llu", rate);
			safe_addstr(win, 0, " ");
			safe_addstr(win, ATTR(CONN_RATE_UNIT), "b/s");
		} else if (rate <= 1024*1024) {
			safe_printw(win, ATTR(CONN_RATE_VALUE), "%4llu", rate/1024);
			safe_addstr(win, 0, " ");
			safe_addstr(win, ATTR(CONN_RATE_UNIT), "k/s");
		} else if (rate <= 1024*1024*1024) {
			safe_printw(win, ATTR(CONN_RATE_VALUE), "%4llu", rate/1024/1024);
			safe_addstr(win, 0, " ");
			safe_addstr(win, ATTR(CONN_RATE_UNIT), "M/s");
		} else {
			safe_printw(win, ATTR(CONN_RATE_VALUE), "%4llu", (rate/(1024*1024*1024)) % 10000);
			safe_addstr(win, 0, " ");
			safe_addstr(win, ATTR(CONN_RATE_UNIT), "G/s");
		}
	}
}

static void
put_rate_spec(WINDOW *win, int64_t src_bytes, int64_t dst_bytes, int64_t time)
{
	int64_t bytes;

	if (time == 0 || (src_bytes == 0 && dst_bytes == 0)) {
		bytes = src_bytes;
		safe_addstr(win, ATTR(CONN_DIRECTION), "  ");
	} else if (src_bytes == dst_bytes) {
		bytes = src_bytes;
		safe_addstr(win, ATTR(CONN_DIRECTION), "<>");
	} else if (src_bytes > dst_bytes) {
		bytes = src_bytes;
		safe_addstr(win, ATTR(CONN_DIRECTION), "->");
	} else {
		bytes = dst_bytes;
		safe_addstr(win, ATTR(CONN_DIRECTION), "<-");
	}

	put_single_rate_spec(win, bytes, time);
}

void
display_connection(ConnTrack *ct, WINDOW *win, int y, int x)
{
	ConnTrackData *data = ct->data;
	int64_t time;
	int64_t src_bytes;
	int64_t dst_bytes;
	uint16_t protocol;
	char rate_char;
	int w1;
	int w2;

	time = time_elapsed(ct->prev_stats.time, ct->cur_stats.time);
	src_bytes = ct->cur_stats.rx_bytes - ct->prev_stats.rx_bytes;
	dst_bytes = ct->cur_stats.tx_bytes - ct->prev_stats.tx_bytes;
	rate_char = get_rate_char(time, src_bytes, dst_bytes,
							  &data->rotator, &data->inactivity,
							  (data->print_time != ct->cur_stats.time));
	data->print_time = ct->cur_stats.time;
	protocol = ct->tuple.protocol;
	written = 0;

	w1 = max(23, (COLS-31)/2 + (COLS-31)%2);
	w2 = max(23, (COLS-31)/2);

	wmove(win, y, x);
	safe_addstr(win, ATTR_CONN_PROTO(protocol), get_connection_type(protocol));
	safe_fillx(win, 5);
	put_host(win, data->src_name, data->src_looked, ct->tuple.src_ip,
			 protocol, ct->tuple.ps.tcp_udp.src_port, w1, data->src_port);
	safe_fillx(win, 6+w1);
	put_host(win, data->dst_name, data->dst_looked, ct->tuple.dst_ip,
			 protocol, ct->tuple.ps.tcp_udp.dst_port, w2, data->dst_port);
	safe_fillx(win, 7+w1+w2);
	put_rate_spec(win, src_bytes, dst_bytes, time);
	safe_printw(win, ATTR(CONN_ROTATOR), " %c ", rate_char);
	if (protocol == CM_PROTO_TCP) {
		safe_addstr(win, ATTR(CONN_TCP_STATE), get_connection_state(ct));
	} else if (protocol == CM_PROTO_ICMP) {
		int16_t type = ct->tuple.ps.icmp.type;
	
		if (type >= 0
				&& type < sizeof(icmp_types)/sizeof(char *)
				&& icmp_types[type] != NULL)
			safe_printw(win, ATTR(CONN_ICMP_TYPE), "%s", icmp_types[type]);
		else
			safe_printw(win, ATTR(CONN_ICMP_TYPE), "(type=%d)", type);
	}
	safe_fillx(win, COLS);
}

void
display_totals(ConnMon *cm, WINDOW *win)
{
	int64_t time;
	int64_t src_bytes;
	int64_t dst_bytes;
	char rate_char;

	time = time_elapsed(cm->total_prev_stats.time, cm->total_cur_stats.time);
	src_bytes = cm->total_cur_stats.rx_bytes - cm->total_prev_stats.rx_bytes;
	dst_bytes = cm->total_cur_stats.tx_bytes - cm->total_prev_stats.tx_bytes;
	rate_char = get_rate_char(time, src_bytes, dst_bytes,
							  &total_rotator, &total_inactivity,
							  (total_print_time != cm->total_cur_stats.time));
	total_print_time = cm->total_cur_stats.time;
	written = 0;

	wmove(win, 0, 0);
	safe_fillx(win, 1);
	safe_printw(win, ATTR(TOTAL_CONN_COUNT), "%d", cm->conn_count);
	safe_addstr(win, 0, " ");
	safe_addstr(win, ATTR(TOTAL_CONN_TEXT),
			    cm->conn_count == 1 ? "connection" : "connections");
	safe_fillx(win, max(COLS-37, 40));
	safe_addstr(win, ATTR(CONN_DIRECTION), "<-");
	safe_addstr(win, 0, " ");
	put_single_rate_spec(win, dst_bytes, time);
	safe_addstr(win, 0, " ");
	safe_addstr(win, ATTR(CONN_DIRECTION), "->");
	safe_addstr(win, 0, " ");
	put_single_rate_spec(win, src_bytes, time);
	safe_addstr(win, 0, " ");
	safe_printw(win, ATTR(CONN_ROTATOR), "%c", rate_char);

	wrefresh(win);
}
