#include <limits.h>
#include <curses.h>
#include "common/memory.h"
#include "common/llist.h"
#include "connmon.h"

ConnList *
create_connection_list(ConnTrackState desired_state, char *title)
{
	ConnList *cl = xmalloc(sizeof(ConnList));

	cl->connections = llist_new();
	cl->connection_it = llist_iterator(cl->connections);
	cl->desired_state = desired_state;
	cl->start_pos = 0;
	cl->added = 0;
	cl->redraw_in = TRUE;
	cl->redraw_out = TRUE;
	cl->title = title;
	cl->window = NULL;
	cl->height = 0;	

	return cl;
}

void
free_connection_list(ConnList *cl)
{
	delwin(cl->window);
	llist_iterate(cl->connections, (IteratorFunc) free_connection);
	llist_free(cl->connections);
	free(cl);
}

void
append_connection(ConnList *cl, ConnTrackData *data)
{
	llist_add(cl->connections, data);
	cl->added++;
}

void
clean_connection_list(ConnList *cl)
{
	iterator_restart(cl->connection_it);
	while (iterator_has_next(cl->connection_it)) {
		ConnTrackData *data = iterator_next(cl->connection_it);
		data->pos = -1;
		if (data->state != cl->desired_state)
			iterator_remove(cl->connection_it);
		if (data->state == CTD_STATE_REMOVED)
			free_connection(data);
	}
}

void
modify_connection_list_bounds(ConnList *cl, int height_mod, int y_mod)
{
	if (y_mod != 0) {
		cl->y += y_mod;
		mvwin(cl->window, cl->y, 0);
	}
	if (height_mod != 0) {
		cl->height += height_mod;
		if (cl->height != 0)
			wresize(cl->window, cl->height, getmaxx(cl->window));
	}
}

void
scroll_connection_list(ConnList *cl, int amount)
{
	int old_start_pos = cl->start_pos;

	if (amount == INT_MAX || amount == INT_MIN) {
		cl->start_pos = amount;
	} else {
		cl->start_pos += amount;
	}
	cl->start_pos = max(cl->start_pos, 0);
	cl->start_pos = min(cl->start_pos, llist_size(cl->connections)-cl->added-cl->height);

	if (cl->start_pos != old_start_pos) {
		cl->redraw_in = TRUE;
		cl->redraw_out = TRUE;
	}
}

void
redisplay_connection_safe(ConnList *cl, ConnTrackData *data)
{
	if (cl->start_pos <= data->pos && data->pos < cl->start_pos
			+ min(llist_size(cl->connections), cl->height)) {
		display_connection(data->ct, cl->window, data->pos - cl->start_pos, 0);
		wrefresh(cl->window);
	}
}

void
redraw_connection_list(ConnList *cl)
{
	int old_conn_count;
	int conn_count;
	int end_pos;
	int pos = 0;

	if (!cl->redraw_in)
		return;

	old_conn_count = llist_size(cl->connections) - cl->added;
	clean_connection_list(cl);
	conn_count = llist_size(cl->connections);
	cl->redraw_in = FALSE;

	if (cl->height == 0)
		return;

	cl->start_pos = min(cl->start_pos, conn_count - getmaxy(cl->window));
	cl->start_pos = max(cl->start_pos, 0);
	cl->added = 0;
	end_pos = cl->start_pos + cl->height;

	iterator_restart(cl->connection_it);
	for (; pos < end_pos && iterator_has_next(cl->connection_it); pos++) {
		ConnTrackData *data = iterator_next(cl->connection_it);
		if (cl->start_pos <= pos) {
			data->pos = pos - cl->start_pos;
			display_connection(data->ct, cl->window, pos-cl->start_pos, 0);
		}
	}

	for (; pos < end_pos && pos < old_conn_count; pos++) {
		if (cl->start_pos <= pos)
			clear_line(cl->window, pos-cl->start_pos);
	}

	wrefresh(cl->window);
}
