#include <config.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include "gconnmon.h"
#include "common/strbuf.h"
#include "common/time64.h"

#define HIPQUAD(addr) \
	((unsigned char *) &addr)[3], \
	((unsigned char *) &addr)[2], \
	((unsigned char *) &addr)[1], \
	((unsigned char *) &addr)[0]

static int64_t
time_elapsed(int64_t time1, int64_t time2)
{
	return (time1 != 0 && time2 != 0 && time1 < time2 ? time2 - time1 : 0);
}

static char *
get_single_rate_spec(int64_t bytes, int64_t time)
{
	static char rate_spec[9];
	int64_t rate;

	if (time == 0)
		return "  -- b/s";

	rate = bytes * MICROS_PER_SEC/time;
	if (rate <= 1024) {
		sprintf(rate_spec, "%4llu b/s", rate);
	} else if (rate <= 1024*1024) {
		sprintf(rate_spec, "%4llu k/s", rate/1024);
	} else if (rate <= 1024*1024*1024) {
		sprintf(rate_spec, "%4llu M/s", rate/1024/1024);
	} else {
		sprintf(rate_spec, "%4llu G/s", (rate/(1024*1024*1024)) % 10000);
	}

	return rate_spec;
}

static char *
get_rate_spec(int64_t src_bytes, int64_t dst_bytes, int64_t time)
{
	int64_t bytes;

	if (time == 0 || (src_bytes == 0 && dst_bytes == 0)) {
		bytes = src_bytes;
//		return "  ";
	} else if (src_bytes == dst_bytes) {
		bytes = src_bytes;
//		return "<>";
	} else if (src_bytes > dst_bytes) {
		bytes = src_bytes;
//		return "->";
	} else {
		bytes = dst_bytes;
//		return "<-";
	}

	return get_single_rate_spec(bytes, time);
}

static char *
get_connection_type(uint16_t protocol)
{
	if (protocol == CM_PROTO_TCP)
		return "tcp";
	if (protocol == CM_PROTO_UDP)
		return "udp";
	if (protocol == CM_PROTO_ICMP)
		return "icmp";
	return "ip";
}

static void
participant_to_string(char **buf, const char *name, uint32_t ip, bool looked,
		int protocol, const char *portname, uint16_t port)
{
	strbuf_clear(buf);

	if (name != NULL) {
		strbuf_append(buf, name);
	} else if (looked) {
		strbuf_appendf(buf, "%u.%u.%u.%u", HIPQUAD(ip));
	} else {
		strbuf_appendf(buf, "(%u.%u.%u.%u)", HIPQUAD(ip));
	}

	if (protocol == CM_PROTO_TCP || protocol == CM_PROTO_UDP) {
		strbuf_append(buf, ":");
		if (portname != NULL) {
			strbuf_append(buf, portname);
		} else {
			strbuf_appendf(buf, "%u", port);
		}
	}
}

static char *
get_connection_state(ConnTrackData *data)
{
	if (data->ct->tuple.protocol == CM_PROTO_TCP) {
		switch (data->ct->tcp_state) {
		case CM_TCP_STATE_ESTABLISHED:		return "ESTABLISHED";
		case CM_TCP_STATE_SYN_SENT:			return "SYN_SENT";
		case CM_TCP_STATE_SYN_RECV:			return "SYN_RECV";
		case CM_TCP_STATE_FIN_WAIT:			return "FIN_WAIT";
		case CM_TCP_STATE_TIME_WAIT:		return "TIME_WAIT";
		case CM_TCP_STATE_CLOSE:			return "CLOSE";
		case CM_TCP_STATE_CLOSE_WAIT:		return "CLOSE_WAIT";
		case CM_TCP_STATE_LAST_ACK:			return "LAST_ACK";
		case CM_TCP_STATE_LISTEN:			return "LISTEN";
		case CM_TCP_STATE_NONE:				return "NONE";
		default:							return "(unknown)";
		}
	}
	else if (data->ct->tuple.protocol == CM_PROTO_ICMP) {
		switch (data->ct->tuple.ps.icmp.type) {
		case 0: 	return "ECHOREPLY";
		case 3: 	return "DST_UNREACH";
		case 4:		return "SRC_QUENCH";
		case 5:		return "REDIRECT";
		case 8:		return "ECHO";
		case 11:	return "TIME_EXCEED";
		case 12:	return "PARAM_PROB";
		case 13:	return "TIMESTAMP";
		case 14:	return "TS_REPLY";
		case 15:	return "INFO_REQ";
		case 16:	return "INFO_REPLY";
		case 17:	return "ADDRESS";
		case 18:	return "ADDR_REPLY";
		default:	return "(unknown)";
		}
	}

	return "";
}

char **
get_connection_columns(ConnTrackData *data)
{
	static char *text[5];
	int64_t time;
	int64_t src_bytes;
	int64_t dst_bytes;

	time = time_elapsed(data->ct->prev_stats.time, data->ct->cur_stats.time);
	src_bytes = data->ct->cur_stats.rx_bytes - data->ct->prev_stats.rx_bytes;
	dst_bytes = data->ct->cur_stats.tx_bytes - data->ct->prev_stats.tx_bytes;

	participant_to_string(&data->src_spec, data->src_host_name,
			data->ct->tuple.src_ip, data->src_host_looked,
			data->ct->tuple.protocol, data->src_port_name,
			data->ct->tuple.ps.tcp_udp.src_port);
	participant_to_string(&data->dst_spec, data->dst_host_name,
			data->ct->tuple.dst_ip, data->dst_host_looked,
			data->ct->tuple.protocol, data->dst_port_name,
			data->ct->tuple.ps.tcp_udp.dst_port);

	text[0] = get_connection_type(data->ct->tuple.protocol);
	text[1] = data->src_spec;
	text[2] = data->dst_spec;
	text[3] = get_rate_spec(src_bytes, dst_bytes, time);
	text[4] = get_connection_state(data);

	return text;
}
