#include <config.h>
#include <gtk/gtk.h>
#include <libconnmon.h>
#include <libhnl.h>
#include <libpnl.h>
#include <stdio.h>
#include "gconnmon.h"
#include "common/error.h"
#include "common/memory.h"
#include "common/time64.h"
#include "common/strbuf.h"
#include "common/llist.h"

gint timeout_tag = -1;
static gint hnl_timeout_tag = -1;
ConnMonSettings cur_settings;
static PNLManager *pnl = NULL;
static HNLManager *hnl = NULL;
static ConnMon *cm = NULL;
static LList *monitor_tags;
static bool any_created;

static void add_hnl_input_monitors(void);
static void hnl_input_occured(gpointer data, gint fd, GdkInputCondition condition);

HNLServiceSpec *
make_hnl_service_spec(const char *name, const char *param)
{
	HNLServiceSpec *spec = xmalloc(sizeof(HNLServiceSpec));
	spec->name = xstrdup(name);
	spec->param = xstrdup(param);
	return spec;
}

void
free_hnl_service_spec(HNLServiceSpec *spec)
{
	free(spec->name);
	free(spec->param);
	free(spec);
}

static int
hnl_timeout_occured(gpointer data)
{
	SelectData sd;

	hnl_clear_select_data(&sd);
	hnl_postselect(hnl, &sd);
	add_hnl_input_monitors();

	return FALSE;
}

static void
hnl_input_occured(gpointer data, gint fd, GdkInputCondition condition)
{
	SelectData sd;

	hnl_clear_select_data(&sd);
	sd.max_fd = fd+1;

	if (condition & GDK_INPUT_READ) {
		//printf("read trigger on fd %d\n", fd);
		FD_SET(fd, &sd.read_fds);
	}
	if (condition & GDK_INPUT_WRITE) {
		//printf("write trigger on fd %d\n", fd);
		FD_SET(fd, &sd.write_fds);
	}
	if (condition & GDK_INPUT_EXCEPTION) {
		//printf("except trigger on fd %d\n", fd);
		FD_SET(fd, &sd.except_fds);
	}
	hnl_postselect(hnl, &sd);

	add_hnl_input_monitors();
}

static void
remove_hnl_input_monitors(void)
{
	while (!llist_is_empty(monitor_tags)) {
		int tag = (int) llist_remove_first(monitor_tags);
		//printf("remove monitor tag %d\n", tag);
		gdk_input_remove(tag);
	}
	if (hnl_timeout_tag != -1)
		gtk_timeout_remove(hnl_timeout_tag);
}

static void
add_hnl_input_monitors(void)
{
	SelectData sd;
	int c;

	remove_hnl_input_monitors();

	hnl_clear_select_data(&sd);
	hnl_preselect(hnl, &sd);

	if (time64_from_timeval(&sd.timeout) != 0) {
		gint32 sleep_time = time64_from_timeval(&sd.timeout) / 1000;
		hnl_timeout_tag = gtk_timeout_add(sleep_time, hnl_timeout_occured, NULL);
	}

	for (c = 0; c < sd.max_fd; c++) {
		if (FD_ISSET(c, &sd.read_fds)) {
			int tag = gdk_input_add(c, GDK_INPUT_READ, hnl_input_occured, NULL);
			llist_add(monitor_tags, (void *) tag);
			//printf("add read monitor on fd %d => tag %d\n", c, tag);
		}
		if (FD_ISSET(c, &sd.write_fds)) {
			int tag = gdk_input_add(c, GDK_INPUT_WRITE, hnl_input_occured, NULL);
			llist_add(monitor_tags, (void *) tag);
			//printf("add write monitor on fd %d => tag %d\n", c, tag);
		}
		if (FD_ISSET(c, &sd.except_fds)) {
			int tag = gdk_input_add(c, GDK_INPUT_EXCEPTION, hnl_input_occured, NULL);
			llist_add(monitor_tags, (void *) tag);
			//printf("add except monitor on fd %d => tag %d\n", c, tag);
		}
	}
}

static bool
is_connection_open(ConnTrack *ct)
{
	return ct->tuple.protocol != CM_PROTO_TCP
			|| ct->tcp_state == CM_TCP_STATE_ESTABLISHED
			|| ct->tcp_state == CM_TCP_STATE_SYN_SENT
			|| ct->tcp_state == CM_TCP_STATE_SYN_RECV
			|| ct->tcp_state == CM_TCP_STATE_LISTEN;
}

static void
connection_created(ConnTrack *ct)
{
	ConnTrackData *data = xmalloc(sizeof(ConnTrackData));

	any_created = TRUE;
	ct->data = data;
	data->ct = ct;
	data->pos = -1;
	data->state = CTD_STATE_CREATED;
	data->src_host_looked = !cur_settings.enable_hnl;
	data->dst_host_looked = !cur_settings.enable_hnl;
	data->src_host_name = NULL;
	data->dst_host_name = NULL;
	if (cur_settings.enable_pnl) {
		data->src_port_name = pnl_lookup(pnl, ct->tuple.protocol, ct->tuple.ps.tcp_udp.src_port);
		data->dst_port_name = pnl_lookup(pnl, ct->tuple.protocol, ct->tuple.ps.tcp_udp.dst_port);
	} else {
		data->src_port_name = NULL;
		data->dst_port_name = NULL;
	}
	data->src_spec = strbuf_new();
	data->dst_spec = strbuf_new();

	if (cur_settings.enable_hnl) {
		data->src_host_query = hnl_lookup(hnl, ct->tuple.src_ip, data);
		if (ct->tuple.src_ip != ct->tuple.dst_ip) {
			data->dst_host_query = hnl_lookup(hnl, ct->tuple.dst_ip, data);
		} else {
			data->dst_host_query = data->src_host_query;
		}
	}
}

static void
lookup_finished(HNLManager *hm, uint32_t ip, char *result, void *userdata)
{
	ConnTrackData *data = userdata;

	if (ip == data->ct->tuple.src_ip) {
		data->src_host_name = (result == NULL ? NULL : xstrdup(result));
		data->src_host_query = NULL;
		data->src_host_looked = TRUE;
	}
	if (ip == data->ct->tuple.dst_ip) {
		data->dst_host_name = (result == NULL ? NULL : xstrdup(result));
		data->dst_host_query = NULL;
		data->dst_host_looked = TRUE;
	}

	refresh_connection(data);
}

static void
connection_updated(ConnTrack *ct)
{
	ConnTrackData *data = ct->data;

	if (is_connection_open(ct)) {
		if (data->state != CTD_STATE_OPEN) {
			data->state = CTD_STATE_OPEN;
			append_connection(data);
		}
	} else {
		if (data->state != CTD_STATE_CLOSED) {
			data->state = CTD_STATE_CLOSED;
			append_connection(data);
		}
	}
}

static void
connection_deleted(ConnTrack *ct)
{
	ConnTrackData *data = ct->data;
	data->state = CTD_STATE_REMOVED;
}

static int
timeout_occured(gpointer data)
{
	any_created = FALSE;
	if (!cm_read_conns(cm))
		die_error();
	if (any_created && cur_settings.enable_hnl)
		add_hnl_input_monitors();
	refresh_connection_lists();
	return TRUE;
}

void
free_connection(ConnTrackData *data)
{
	if (data->dst_host_query != NULL && data->src_host_query != data->dst_host_query) {
		hnl_cancel(hnl, data->dst_host_query);
		data->dst_host_query = NULL;
	}
	if (data->src_host_query != NULL) {
		hnl_cancel(hnl, data->src_host_query);
		data->src_host_query = NULL;
	}
	if (data->src_host_name != NULL)
		free(data->src_host_name);
	if (data->dst_host_name != NULL)
		free(data->dst_host_name);

	strbuf_free(&data->src_spec);
	strbuf_free(&data->dst_spec);
	
	free(data);
}

void
free_settings(ConnMonSettings *s)
{
	free(s->driver_name);
	free(s->driver_param);
	free(s->filter);
	llist_iterate(s->hnl_services, (IteratorFunc) free_hnl_service_spec);
	llist_clear(s->hnl_services);
	llist_iterate(s->pnl_files, free);
	llist_clear(s->pnl_files);
}

void
activate_settings(ConnMonSettings *s)
{
	Iterator *it;
	char *driver;
	char *hnl_services;
	char *pnl_files;

	remove_hnl_input_monitors();

	driver = strbuf_new();
	strbuf_append(&driver, s->driver_name);
	if (s->driver_param != NULL && strcmp(s->driver_param, "") != 0)
		strbuf_appendf(&driver, ":%s", s->driver_param);
	hnl_services = strbuf_new();
	for (it = llist_iterator(s->hnl_services); iterator_has_next(it); ) {
		HNLServiceSpec *spec = iterator_next(it);
		if (hnl_services[0] != '\0')
			strbuf_append(&hnl_services, ",");
		strbuf_append(&hnl_services, spec->name);
		if (spec->param != NULL && strcmp(spec->param, "") != 0)
			strbuf_appendf(&driver, ":%s", spec->param);
	}
	iterator_free(it);
	pnl_files = strbuf_new();
	for (it = llist_iterator(s->pnl_files); iterator_has_next(it); ) {
		char *file = iterator_next(it);
		if (pnl_files[0] != '\0')
			strbuf_append(&pnl_files, ",");
		strbuf_append(&pnl_files, file);
	}
	iterator_free(it);

	if (pnl != NULL)
		pnl_free(pnl);
	if (hnl != NULL)
		hnl_free(hnl);
	if (cm != NULL)
		cm_free(cm);
	if (timeout_tag != -1)
		gtk_input_remove(timeout_tag);

	if (s->enable_pnl) {
		pnl = pnl_new();
		if (!pnl_initialize_from_spec(pnl, pnl_files))
			die_error();
	} else {
		pnl = NULL;
	}

	if (s->enable_hnl) {
		hnl = hnl_new();
		hnl->lookup_finished = lookup_finished;
		hnl_set_caching(hnl, s->cache_hnl);
		if (!hnl_initialize_from_spec(hnl, hnl_services))
			die_error();
	} else {
		hnl = NULL;
	}

	cm = cm_new(driver);
	if (cm == NULL)
		die_error();
	cm->create_handler = connection_created;
	cm->delete_handler = connection_deleted;
	cm->update_handler = connection_updated;

	strbuf_free(&driver);
	strbuf_free(&hnl_services);
	strbuf_free(&pnl_files);

	reset_everything();
	if (!cm_read_conns(cm))
		die_error();
	if (cur_settings.enable_hnl)
		add_hnl_input_monitors();
	refresh_connection_lists();
	timeout_tag = gtk_timeout_add(cur_settings.poll_delay, timeout_occured, NULL);
}

int
main(int argc, char **argv)
{
	monitor_tags = llist_new();

	cur_settings.driver_name = xstrdup("netfilter");
	cur_settings.driver_param = xstrdup("");
	cur_settings.filter = xstrdup("");
	cur_settings.poll_delay = DEFAULT_SLEEP_TIME_MS;
	cur_settings.enable_hnl = TRUE;
	cur_settings.cache_hnl = TRUE;
	cur_settings.enable_pnl = TRUE;
	cur_settings.hnl_services = llist_new();
	llist_add(cur_settings.hnl_services, make_hnl_service_spec("hosts", ""));
	llist_add(cur_settings.hnl_services, make_hnl_service_spec("adns", ""));
	cur_settings.pnl_files = llist_new();
	llist_add(cur_settings.pnl_files, xstrdup("/etc/services"));

	gtk_init(&argc, &argv);
	create_main_window();

	activate_settings(&cur_settings);

	gtk_main();

	free_settings(&cur_settings);
	llist_free(monitor_tags);

	if (pnl != NULL)
		pnl_free(pnl);
	if (hnl != NULL)
		hnl_free(hnl);
	if (cm != NULL)
		cm_free(cm);

	exit(0);
}
