#include <config.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include "common/error.h"
#include "common/memory.h"
#include "common/string-utils.h"
#include "libconnmon-internal.h"

#define LINUX_TCP_FILE "/proc/net/tcp"
#define LINUX_UDP_FILE "/proc/net/udp"

typedef struct {
	ConnMonPrivate cmp;
	FILE *tcp_file;		/* file which tcp connections are read from */
	char *tcp_name;		/* name of tcp file */
	FILE *udp_file;		/* file which udp connections are read from */
	char *udp_name;		/* name of udp file */
	char *linebuf;		/* buffer we read lines into */
	int linebuf_size;	/* size of buffer */
} LinuxConnMon;

static bool linux_cm_initialize(ConnMonPrivate *cmp, const char *args);
static void linux_cm_free(ConnMonPrivate *cmp);
static bool linux_cm_read_conns(ConnMonPrivate *cmp, int64_t time);

ConnMonDriver linux_cm_driver = {
	sizeof(LinuxConnMon),
	linux_cm_initialize,
	linux_cm_free,
	linux_cm_read_conns,
};

/* Utility function to report an error where errno contains
 * the error code.
 */
static void
set_errno_error(const char *name, const char *msg)
{
	set_error("%s `%s': %s", msg, name, strerror(errno));
}

static bool
linux_cm_initialize(ConnMonPrivate *cmp, const char *args)
{
	LinuxConnMon *lcm = (LinuxConnMon *) cmp;

	lcm->linebuf = NULL;
	lcm->linebuf_size = 0;
	lcm->tcp_name = xstrdup(LINUX_TCP_FILE);
	lcm->udp_name = xstrdup(LINUX_UDP_FILE);
	lcm->tcp_file = NULL;
	lcm->udp_file = NULL;

	lcm->tcp_file = fopen(lcm->tcp_name, "r");
	if (lcm->tcp_file == NULL) {
		set_errno_error(lcm->tcp_name, "cannot open");
		return FALSE;
	}

	lcm->udp_file = fopen(lcm->udp_name, "r");
	if (lcm->udp_file == NULL) {
		set_errno_error(lcm->udp_name, "cannot open");
		return FALSE;
	}

	return TRUE;
}

static void
linux_cm_free(ConnMonPrivate *cmp)
{
	LinuxConnMon *lcm = (LinuxConnMon *) cmp;

	free(lcm->tcp_name);
	free(lcm->udp_name);
	if (lcm->tcp_file != NULL)
		fclose(lcm->tcp_file);
	if (lcm->udp_file != NULL)
		fclose(lcm->udp_file);
	if (lcm->linebuf != NULL)
		free(lcm->linebuf);
}

static bool
process_tcp_line(LinuxConnMon *lcm, char *line, int64_t time)
{
	ConnTrack *ct;
	ConnTuple tuple;
	ConnStats stats;
	int count;
	uint16_t state;
	uint32_t timeout;

	count = strlen(line);
	if (count != 127 && count != 149) {
		set_error("%s: cannot parse the following line (length != 127/149):\n%s",
				lcm->tcp_name, line);
		return FALSE;
	}

	memclear(&stats, sizeof(ConnStats));

	// FIXME: there are probably other fields we want to parse, but since
	// FIXME: there's no support for displaying that, we don't care about them
	count = sscanf(line,
			"%*d: %8X:%4hX %8X:%4hX %2hx %*X:%*X %*X:%*X %*X %*d %8X",
			&tuple.src_ip, &tuple.ps.tcp_udp.src_port,
			&tuple.dst_ip, &tuple.ps.tcp_udp.dst_port,
			&state, &timeout);
	if (count != 6) {
		set_error("%s: cannot parse the following line:\n%s",
				lcm->tcp_name, line);
		return FALSE;
	}

	tuple.protocol = CM_PROTO_TCP;
	tuple.src_ip = ntohl(tuple.src_ip);
	tuple.dst_ip = ntohl(tuple.dst_ip);
	tuple.ps.tcp_udp.src_port = ntohs(tuple.ps.tcp_udp.src_port);
	tuple.ps.tcp_udp.dst_port = ntohs(tuple.ps.tcp_udp.dst_port);

	/* FIXME: make an option to include/exclude these */
	if (state != CM_TCP_STATE_LISTEN) {
		ct = cm_ct_lookup(&lcm->cmp, &tuple, &stats);
		if (ct != NULL) {
			ct->tcp_state = state;
			ct->timeout = timeout;
			cm_ct_report(&lcm->cmp, ct);
		}
	}

	return TRUE;
}

static bool
process_udp_line(LinuxConnMon *lcm, char *line, int64_t time)
{
	ConnTrack *ct;
	ConnTuple tuple;
	ConnStats stats;
	int count;
	uint16_t state;
	uint32_t timeout;

	if (strlen(line) != 127) {
		set_error("%s: cannot parse the following line (length != 127):\n%s",
				lcm->udp_name, line);
		return FALSE;
	}

	memclear(&stats, sizeof(ConnStats));

	// FIXME: there are probably other fields we want to parse, but since
	// FIXME: there's no support for displaying that, we don't care about them
	count = sscanf(line,
			"%*d: %8X:%4hX %8X:%4hX %2hx %*X:%*X %*X:%*X %*X %*d %8X",
			&tuple.src_ip, &tuple.ps.tcp_udp.src_port,
			&tuple.dst_ip, &tuple.ps.tcp_udp.dst_port,
			&state, &timeout);
	if (count != 6) {
		set_error("%s: cannot parse the following line:\n%s", lcm->udp_name, line);
		return FALSE;
	}

	tuple.protocol = CM_PROTO_UDP;
	tuple.src_ip = ntohl(tuple.src_ip);
	tuple.dst_ip = ntohl(tuple.dst_ip);
	tuple.ps.tcp_udp.src_port = ntohs(tuple.ps.tcp_udp.src_port);
	tuple.ps.tcp_udp.dst_port = ntohs(tuple.ps.tcp_udp.dst_port);

	/* FIXME: make an option to include/exclude these */
	if (state == CM_TCP_STATE_ESTABLISHED) {
		ct = cm_ct_lookup(&lcm->cmp, &tuple, &stats);
		if (ct != NULL) {
			ct->tcp_state = state;
			ct->timeout = timeout;
			cm_ct_report(&lcm->cmp, ct);
		}
	}

	return TRUE;
}

static bool
linux_cm_read_conns(ConnMonPrivate *cmp, int64_t time)
{
	LinuxConnMon *lcm = (LinuxConnMon *) cmp;
	bool first;

	first = TRUE;
	while (getline(&lcm->linebuf, &lcm->linebuf_size, lcm->tcp_file) != -1) {
		chomp(lcm->linebuf);
		if (!first && !process_tcp_line(lcm, lcm->linebuf, time))
			return FALSE;
		first = FALSE;
	}

	if (ferror(lcm->tcp_file)) {
		set_errno_error(lcm->tcp_name, "cannot read");
		return FALSE;
	}
	if (fseek(lcm->tcp_file, 0, SEEK_SET) < 0) {
		set_errno_error(lcm->tcp_name, "cannot seek in");
		return FALSE;
	}

	first = TRUE;
	while (getline(&lcm->linebuf, &lcm->linebuf_size, lcm->udp_file) != -1) {
		chomp(lcm->linebuf);
		if (!first && !process_udp_line(lcm, lcm->linebuf, time))
			return FALSE;
		first = FALSE;
	}

	if (ferror(lcm->udp_file)) {
		set_errno_error(lcm->udp_name, "cannot read");
		return FALSE;
	}
	if (fseek(lcm->udp_file, 0, SEEK_SET) < 0) {
		set_errno_error(lcm->udp_name, "cannot seek in");
		return FALSE;
	}

	return TRUE;
}
