#include <config.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "common/error.h"
#include "common/memory.h"
#include "common/intparse.h"
#include "common/string-utils.h"
#include "libconnmon-internal.h"

#define IP_CONNTRACK_FILE "/proc/net/ip_conntrack"

typedef struct {
	ConnMonPrivate cmp;
	FILE *file;			/* file which connections are read from */
	char *name;			/* name of file */
	char *linebuf;		/* buffer we read lines into */
	int linebuf_size;	/* size of buffer */
} NetfilterConnMon;

static bool netfilter_initialize(ConnMonPrivate *cmp, const char *args);
static void netfilter_free(ConnMonPrivate *cmp);
static bool netfilter_read_conns(ConnMonPrivate *cmp, int64_t time);

ConnMonDriver netfilter_cm_driver = {
	sizeof(NetfilterConnMon),
	netfilter_initialize,
	netfilter_free,
	netfilter_read_conns,
};

/* Utility function to report an error where errno contains
 * the error code.
 */
static void
set_errno_error(NetfilterConnMon *ncm, const char *msg)
{
	set_error("%s `%s': %s", msg, ncm->name, strerror(errno));
}

static bool
expect(char **line, char *str)
{
	int len = strlen(str);
	if (strncmp(*line, str, len) != 0)
		return FALSE;
	*line += len;
	return TRUE;
}

static bool
expect_n(char **line, int count)
{
	char *cur = *line;
	for (; count > 0; count--) {
		if (*cur == '\0')
			return FALSE;
		cur++;
	}
	*line = cur;
	return TRUE;
}

static bool
read_until(char **line, char until, char **out)
{
	char *cur = *line;
	*out = cur;
	while (*cur && *cur != until)
		cur++;
	if (*out == cur)
		return FALSE;
	*cur = '\0';
	*line = cur+1;
	return TRUE;
}

static bool
read_ip(char **line, uint32_t *ip)
{
	char *tok;
	uint8_t a, b, c, d;

	if (!read_until(line, '.', &tok) || !parse_uint8(tok, &a)
			|| !read_until(line, '.', &tok) || !parse_uint8(tok, &b)
			|| !read_until(line, '.', &tok) || !parse_uint8(tok, &c)
			|| !read_until(line, ' ', &tok) || !parse_uint8(tok, &d))
		return FALSE;

	*ip = (a << 24) | (b << 16) | (c << 8) | d;
	return TRUE;
}

static bool
parse_tuple(char **line, ConnTuple *ct, uint16_t protocol)
{
	char *tok;

	if (!expect(line, "src=")
			|| !read_ip(line, &ct->src_ip)
			|| !expect(line, "dst=")
			|| !read_ip(line, &ct->dst_ip))
		return FALSE;

	if (protocol == CM_PROTO_TCP || protocol == CM_PROTO_UDP)
		return expect(line, "sport=")
				&& read_until(line, ' ', &tok)
				&& parse_uint16(tok, &ct->ps.tcp_udp.src_port)
				&& expect(line, "dport=")
				&& read_until(line, ' ', &tok)
				&& parse_uint16(tok, &ct->ps.tcp_udp.dst_port);

	if (protocol == CM_PROTO_ICMP)
		return expect(line, "type=")
				&& read_until(line, ' ', &tok)
				&& parse_uint8(tok, &ct->ps.icmp.type)
				&& expect(line, "code=")
				&& read_until(line, ' ', &tok)
				&& parse_uint8(tok, &ct->ps.icmp.code)
				&& expect(line, "id=")
				&& read_until(line, ' ', &tok)
				&& parse_uint16(tok, &ct->ps.icmp.id);

	ct->ps.all = 0;
	return TRUE;
}

static bool
parse_tcp_state(char *line, uint8_t *status)
{
	if (strcmp(line, "NONE") == 0)
		*status = CM_TCP_STATE_NONE;
	else if (strcmp(line, "ESTABLISHED") == 0)
		*status = CM_TCP_STATE_ESTABLISHED;
	else if (strcmp(line, "SYN_SENT") == 0)
		*status = CM_TCP_STATE_SYN_SENT;
	else if (strcmp(line, "SYN_RECV") == 0)
		*status = CM_TCP_STATE_SYN_RECV;
	else if (strcmp(line, "FIN_WAIT") == 0)
		*status = CM_TCP_STATE_FIN_WAIT;
	else if (strcmp(line, "TIME_WAIT") == 0)
		*status = CM_TCP_STATE_TIME_WAIT;
	else if (strcmp(line, "CLOSE") == 0)
		*status = CM_TCP_STATE_CLOSE;
	else if (strcmp(line, "CLOSE_WAIT") == 0)
		*status = CM_TCP_STATE_CLOSE_WAIT;
	else if (strcmp(line, "LAST_ACK") == 0)
		*status = CM_TCP_STATE_LAST_ACK;
	else if (strcmp(line, "LISTEN") == 0)
		*status = CM_TCP_STATE_LISTEN;
	else
		return FALSE;

	return TRUE;
}

static bool
netfilter_initialize(ConnMonPrivate *cmp, const char *args)
{
	NetfilterConnMon *ncm = (NetfilterConnMon *) cmp;

	ncm->linebuf = NULL;
	ncm->linebuf_size = 0;
	if (args != NULL && *args != '\0')
		ncm->name = xstrdup(args);
	else
		ncm->name = xstrdup(IP_CONNTRACK_FILE);
	ncm->file = fopen(ncm->name, "r");

	if (ncm->file == NULL) {
		set_errno_error(ncm, "cannot open");
		return FALSE;
	}

	return TRUE;
}

static void
netfilter_free(ConnMonPrivate *cmp)
{
	NetfilterConnMon *ncm = (NetfilterConnMon *) cmp;

	free(ncm->name);
	if (ncm->file != NULL)
		fclose(ncm->file);
	if (ncm->linebuf != NULL)
		free(ncm->linebuf);
}

static bool
netfilter_read_conns(ConnMonPrivate *cmp, int64_t time)
{
	NetfilterConnMon *ncm = (NetfilterConnMon *) cmp;

	while (getline(&ncm->linebuf, &ncm->linebuf_size, ncm->file) != -1) {
		ConnTuple original_tuple;
		ConnTuple reply_tuple;
		ConnStats stats;
		ConnTrack *ct;
		char *line;
		char *tok;
		uint32_t timeout;
		uint32_t use_count;
		uint16_t protocol;
		uint8_t tcp_state;
		uint8_t flags;

		line = ncm->linebuf;
		chomp(line);

		if (starts_with(line, "EXPECTING: "))
			continue;
		if (!expect_n(&line, 9)
				|| !read_until(&line, ' ', &tok)
				|| !parse_uint16(tok, &protocol)
				|| !read_until(&line, ' ', &tok)
				|| !parse_uint32(tok, &timeout)) {
			set_error("%s: cannot parse the following line:\n%s", ncm->name, line);
			return FALSE;
		}
		if (protocol == CM_PROTO_TCP) {
			if (!read_until(&line, ' ', &tok)
					|| !parse_tcp_state(tok, &tcp_state)) {
				set_error("%s: cannot parse the following line:\n%s", ncm->name, line);
				return FALSE;
			}
		} else {
			tcp_state = 0;
		}

		flags = 0;
		if (!parse_tuple(&line, &original_tuple, protocol)) {
			set_error("%s: cannot parse line:\n%s", ncm->name, line);
			return FALSE;
		}
		original_tuple.protocol = protocol;

		if (expect(&line, "[UNREPLIED] "))
			flags |= CM_CT_UNREPLIED;
		if (!parse_tuple(&line, &reply_tuple, protocol)) {
			set_error("%s: cannot parse line:\n%s", ncm->name, line);
			return FALSE;
		}
		if (expect(&line, "[ASSURED] "))
			flags |= CM_CT_ASSURED;
		if (!expect(&line, "use=")
				|| !read_until(&line, ' ', &tok)
				|| !parse_uint32(tok, &use_count)) {
			set_error("%s: cannot parse line:\n%s", ncm->name, line);
			return FALSE;
		}
		if (!expect(&line, "rx=")
				|| !read_until(&line, ' ', &tok)
				|| !parse_uint64(tok, &stats.rx_packets)
				|| !read_until(&line, ' ', &tok)
				|| !parse_uint64(tok, &stats.rx_bytes)
				|| !expect(&line, "tx=")
				|| !read_until(&line, ' ', &tok)
				|| !parse_uint64(tok, &stats.tx_packets)
				|| !read_until(&line, ' ', &tok)
				|| !parse_uint64(tok, &stats.tx_bytes)) {
			memset(&stats, 0, sizeof(ConnStats));
		} else {
			stats.time = time;
		}

		ct = cm_ct_lookup(cmp, &original_tuple, &stats);
		if (ct != NULL) {
			ct->timeout = timeout;
			ct->tcp_state = tcp_state;
			ct->flags |= flags;
			ct->use_count = use_count;
			cm_ct_report(cmp, ct);
		}
	}

	if (ferror(ncm->file)) {
		set_errno_error(ncm, "cannot read ");
		return FALSE;
	}

	if (fseek(ncm->file, 0, SEEK_SET) < 0) {
		set_errno_error(ncm, "cannot seek in ");
		return FALSE;
	}

	return TRUE;
}
