%{

#include <config.h>
#include <stdio.h>
#include <ctype.h>
#include "common/memory.h"
#include "common/strbuf.h"
#include "common/string-utils.h"
#include "common/llist.h"
#include "common/error.h"
#include "filter.h"

typedef struct {
	char *name;
	int token;
} OpInfo;

static const char *const field_names[] = {
	"port", "src_port", "dst_port", "host", "src_host", "dst_host",
	"protocol", "icmp_type", "icmp_code", "icmp_id",
	"tcp", "udp", "icmp", "unknown"
};

static OpInfo op_info[] = {
	{ "and", LOGOP_AND },
	{ "or", LOGOP_OR },
	{ "xor", LOGOP_XOR },
	{ "=", CMPOP_EQ },
	{ "!=", CMPOP_NE },
	{ "<", CMPOP_LT },
	{ ">", CMPOP_GT },
	{ "<=", CMPOP_LE },
	{ ">=", CMPOP_GE },
};

static const char *input;
static int pos;
static LList *string_list;
static Expr *result;

static Expr *new_logical_expr(Operator op, Expr *expr1, Expr *expr2);
static Expr *new_compare_expr(Operator op, FieldType field, char *string);
static Expr *new_not_expr(Expr *expr1);
static int yylex(void);
static void yyerror(char *msg);
static int yyparse (void);

%}

%union {
	FieldType field;
	Expr *expr;
	Operator op;
	char *string;
}
%token FIELD STRING
%left OR XOR
%left AND
%left NOT
%left EQ NE LT GT LE GE
%type <field> FIELD
%type <expr> expr input
%type <op> AND OR XOR EQ NE LT GT LE GE
%type <string> STRING

%%

input:	expr					{ result = $1; }
		;
expr:	'(' expr ')'			{ $$ = $2; }
		| NOT expr				{ $$ = new_not_expr($2); }
		| expr AND expr			{ $$ = new_logical_expr($2, $1, $3); }
		| expr OR expr			{ $$ = new_logical_expr($2, $1, $3); }
		| expr XOR expr			{ $$ = new_logical_expr($2, $1, $3); }
		| FIELD EQ STRING		{ $$ = new_compare_expr($2, $1, $3); }
		| FIELD NE STRING		{ $$ = new_compare_expr($2, $1, $3); }
		| FIELD LT STRING		{ $$ = new_compare_expr($2, $1, $3); }
		| FIELD GT STRING		{ $$ = new_compare_expr($2, $1, $3); }
		| FIELD LE STRING		{ $$ = new_compare_expr($2, $1, $3); }
		| FIELD GE STRING		{ $$ = new_compare_expr($2, $1, $3); }
		;

%%

static Expr *
new_logical_expr(Operator op, Expr *expr1, Expr *expr2)
{
	LogicalExpr *expr = xmalloc(sizeof(LogicalExpr));
	expr->e.type = EXPR_LOGICAL;
	expr->op = op;
	expr->expr1 = expr1;
	expr->expr2 = expr2;
	return &expr->e;
}

static Expr *
new_compare_expr(Operator op, FieldType field, char *string)
{
	CompareExpr *expr = xmalloc(sizeof(CompareExpr));
	expr->e.type = EXPR_COMPARE;
	expr->op = op;
	expr->field = field;
	expr->string = xstrdup(string);
	return &expr->e;
}

static Expr *
new_not_expr(Expr *expr1)
{
	NotExpr *expr = xmalloc(sizeof(NotExpr));
	expr->e.type = EXPR_NOT;
	expr->expr1 = expr1;
	return &expr->e;
}

static int
yylex(void)
{
	int start;
	int c;

	for (; input[pos] && isspace(input[pos]); pos++);
	if (!input[pos])
		return EOF;

	if (input[pos] == '(' || input[pos] == ')')
		return input[pos++];

	if (input[pos] == '"') {
		char *strbuf = strbuf_new();
		for (pos++; input[pos] && input[pos] != '"'; pos++) {
			if (input[pos] == '\\')
				pos++;
			strbuf_append_char(&strbuf, input[pos]);
		}
		if (!input[pos])
			die("unterminated string");
		yylval.string = strbuf_free_to_string(&strbuf);
		llist_add(string_list, yylval.string);
		return STRING;
	}

	start = pos;
	for (; input[pos] && !isspace(input[pos]); pos++);

	for (c = 0; c < sizeof(op_info)/sizeof(OpInfo); c++) {
		if (strncmp(input+start, op_info[c].name, pos-start) == 0) {
			yylval.op = c;
			return op_info[c].token;
		}
	}
	for (c = 0; c < sizeof(field_names)/sizeof(char *); c++) {
		if (strncmp(input+start, field_names[c], pos-start) == 0) {
			yylval.field = c;
			return FIELD;
		}
	}

	yylval.string = substring(input, start, pos);
	llist_add(string_list, yylval.string);
	return STRING;
}

static void
yyerror(char *msg)
{
	die("%s in column %d", msg, pos);
}

Expr *
parse_expression(const char *string)
{
	input = string;
	pos = 0;
	string_list = llist_new();
	result = NULL;

	if (yyparse() != 0)
		exit(1);

	llist_iterate(string_list, free);
	llist_free(string_list);

	return result;
}
