#include <config.h>
#include <stdarg.h>
#include <stdio.h>
#include "common/error.h"
#include "common/memory.h"
#include "common/time64.h"
#include "common/string-utils.h"
#include "libconnmon-internal.h"

static uint32_t hash_tuple(ConnTuple *ct);
static int32_t compare_tuples(ConnTuple *ct1, ConnTuple *ct2);

static uint32_t
hash_tuple(ConnTuple *ct)
{
	return ct->src_ip
			+ ct->dst_ip
			+ ct->ps.tcp_udp.src_port
			+ ct->ps.tcp_udp.dst_port
			+ ct->protocol;
}

static int32_t
compare_tuples(ConnTuple *ct1, ConnTuple *ct2)
{
	RETURN_IF_DIFFERENT(ct1->protocol, ct2->protocol);
	RETURN_IF_DIFFERENT(ct1->src_ip, ct2->src_ip);
	RETURN_IF_DIFFERENT(ct1->dst_ip, ct2->dst_ip);
	RETURN_IF_DIFFERENT(ct1->ps.all, ct2->ps.all);
	return 0;
}

ConnMon *
cm_new(const char *args)
{
	ConnMonDriver *driver;
	ConnMonPrivate *cmp;
	int pos;

	if (args == NULL)
		args = DEFAULT_CONNMON_DRIVER;

	pos = strindex(args, ':');
	if (pos == -1)
		pos = strlen(args);

	if (strncmp(args, "netfilter", pos) == 0) {
		driver = &netfilter_cm_driver;
	} else if (strncmp(args, "linux", pos) == 0) {
		driver = &linux_cm_driver;
	} else {
		char *tmp = substring(args, 0, pos);
		set_error("no such driver `%s'", tmp);
		free(tmp);
		return NULL;
	}

	cmp = xmalloc(driver == NULL ? sizeof(ConnMonPrivate) : driver->struct_size);
	memclear(cmp, sizeof(ConnMonPrivate));
	cmp->driver = driver;
	cmp->connections = hmap_new();
	cmp->it = hmap_value_iterator(cmp->connections);
	cmp->connection_unused = llist_new();
	hmap_set_hash_function(cmp->connections, (HashFunc) hash_tuple);
	hmap_set_compare_function(cmp->connections, (CompareFunc) compare_tuples);

	args = (args[pos] == ':' ? args+pos+1 : NULL);
	if (cmp->driver->initialize != NULL && !cmp->driver->initialize(cmp, args)) {
		cmp->driver = NULL;
		cm_free(&cmp->cm);
		return NULL;
	}

	return &cmp->cm;
}

/* Free this libconnmon engine. This should be called even if
 * an error has occured, and even if cm_new failed (returned FALSE).
 */
void
cm_free(ConnMon *cm)
{
	ConnMonPrivate *cmp = (ConnMonPrivate *) cm;

	if (cmp->driver != NULL && cmp->driver->free != NULL)
		cmp->driver->free(cmp);
	if (cmp->it != NULL) {
		iterator_restart(cmp->it);
		while (iterator_has_next(cmp->it)) {
			ConnTrack *ct = iterator_next(cmp->it);
			//XXX: No point in doing this really...
			//cm->delete_handler(ct);
			free(ct);
		}
		iterator_free(cmp->it);
	}
	if (cmp->connection_unused != NULL) {
		llist_iterate(cmp->connection_unused, free);
		llist_free(cmp->connection_unused);
	}
	if (cmp->connections != NULL)
		hmap_free(cmp->connections);

	free(cmp);
}

/* This function is called by line-based drivers when they
 * encounter a connection. It will return an old or new
 * ConnTrack object, or NULL if this connection was not
 * desired (i.e. if it was filtered out).
 */
ConnTrack *
cm_ct_lookup(ConnMonPrivate *cmp, ConnTuple *tuple, ConnStats *stats)
{
	ConnTrack *ct;

	if (cmp->filter != NULL && !try_expression(cmp->filter, tuple))
		return NULL;

	ct = hmap_get(cmp->connections, tuple);
	if (ct == NULL) {
		if (!llist_is_empty(cmp->connection_unused)) {
			ct = llist_pop(cmp->connection_unused);
		} else {
			ct = xmalloc(sizeof(ConnTrack));
		}
		ct->data = NULL;
		ct->tuple = *tuple;
		ct->first_stats = *stats;
		ct->flags = 0;
		memset(&ct->prev_stats, 0, sizeof(ConnStats));
		hmap_put(cmp->connections, &ct->tuple, ct);
	} else {
		ct->flags = CM_CT_SEEN;
		ct->prev_stats = ct->cur_stats;
	}
	ct->cur_stats = *stats;

	return ct;
}

/* This function is called by line-based drivers when they
 * have filled in all information of a connection that was
 * previously returned by cm_ct_lookup.
 */
void
cm_ct_report(ConnMonPrivate *cmp, ConnTrack *ct)
{
	if (!(ct->flags & CM_CT_SEEN) && cmp->cm.create_handler != NULL)
		cmp->cm.create_handler(ct);

	if (cmp->cm.update_handler != NULL)
		cmp->cm.update_handler(ct);

	if (ct->cur_stats.time != 0) {
		cmp->cm.total_cur_stats.rx_packets += ct->cur_stats.rx_packets - ct->prev_stats.rx_packets;
		cmp->cm.total_cur_stats.tx_packets += ct->cur_stats.tx_packets - ct->prev_stats.tx_packets;
		cmp->cm.total_cur_stats.rx_bytes += ct->cur_stats.rx_bytes - ct->prev_stats.rx_bytes;
		cmp->cm.total_cur_stats.tx_bytes += ct->cur_stats.tx_bytes - ct->prev_stats.tx_bytes;
	}

	cmp->cm.conn_count++;
	ct->flags |= CM_CT_SEEN;
}

/* Read connections, reporting new ones by calling
 * create_handler, updated ones by calling update_handler,
 * and finally those no longer listed by calling delete_handler.
 */
bool
cm_read_conns(ConnMon *cm)
{
	ConnMonPrivate *cmp = (ConnMonPrivate *) cm;
	bool got_stats;
	int64_t time;

	cm->conn_count = 0;
	cm->total_prev_stats = cm->total_cur_stats;
	got_stats = FALSE;
	time = time64_now();

	if (!cmp->driver->read_conns(cmp, time))
		return FALSE;

	iterator_restart(cmp->it);
	while (iterator_has_next(cmp->it)) {
		ConnTrack *ct = iterator_next(cmp->it);

		if ((ct->flags & CM_CT_SEEN) == 0) {
			if (cmp->cm.delete_handler != NULL)
				cmp->cm.delete_handler(ct);
			llist_add(cmp->connection_unused, ct);
			iterator_remove(cmp->it);
		}
		else {
			if (ct->cur_stats.time != 0)
				got_stats = TRUE;
			ct->flags &= ~CM_CT_SEEN;
		}
	}

	if (got_stats) {
		cm->total_cur_stats.time = time;
		if (cm->total_first_stats.time == 0)
			cm->total_first_stats = cm->total_cur_stats;
	}

	return TRUE;
}

/* For each known connection, call update_handler with
 * that connection.
 */
void
cm_get_conns(ConnMon *cm)
{
	ConnMonPrivate *cmp = (ConnMonPrivate *) cm;

	iterator_restart(cmp->it);
	while (iterator_has_next(cmp->it)) {
		ConnTrack *ct = iterator_next(cmp->it);
		cm->update_handler(ct);
	}
}

/* Install or remove a filter that controls what
 * connections are reported. If filter is NULL,
 * the filter is removed.
 */
bool
cm_set_filter(ConnMon *cm, const char *filter)
{
	ConnMonPrivate *cmp = (ConnMonPrivate *) cm;

	if (cmp->filter != NULL)
		free_expression(cmp->filter);
	cmp->filter = parse_expression(filter);
	/* FIXME: error reporting! */

	return TRUE;
}
