#include <config.h>
#include <adns.h>
#include "common/error.h"
#include "common/memory.h"
#include "common/common.h"
#include "libhnl-internal.h"

typedef struct {
	HNLService svc;
	uint32_t lookup_count;
	adns_state adns;
} AdnsHNLService;

typedef struct {
	HNLQuery *query;
	adns_query adns_query;
	bool cancelled;
} AdnsHNLQueryData;

#define IPQUAD(addr) \
	((unsigned char *) &addr)[0], \
	((unsigned char *) &addr)[1], \
	((unsigned char *) &addr)[2], \
	((unsigned char *) &addr)[3]

static void
adns_hnl_free(HNLService *svc)
{
    AdnsHNLService *asvc = (AdnsHNLService *) svc;
	adns_finish(asvc->adns);
	free(asvc);
}

static void
adns_hnl_lookup(HNLService *svc, HNLQuery *query)
{
    AdnsHNLService *asvc = (AdnsHNLService *) svc;
	AdnsHNLQueryData *data = xmalloc(sizeof(AdnsHNLQueryData));
	char ipstr[30];
	int rc;

	data->query = query;
	data->cancelled = FALSE;
	query->servicedata = data;
	sprintf(ipstr, "%d.%d.%d.%d.in-addr.arpa", IPQUAD(query->ip));

	rc = adns_submit(asvc->adns, ipstr, adns_r_ptr, adns_qf_search, data,
			&data->adns_query);
	if (rc != 0)
		die("cannot submit DNS query: %s", strerror(rc));

	asvc->lookup_count++;
}

static void
adns_hnl_cancel(HNLService *svc, HNLQuery *query)
{
	AdnsHNLQueryData *data = query->servicedata;

	/* Cancelling queries resulted in some problems, mainly
	   ../src/check.c:190: adns__consistency: Assertion `( search)' failed.
       and sometimes other crashes. Instead we don't cancel
	   queries, we ignore their results. */
	/* adns_cancel(data->adns_query); */

	data->cancelled = TRUE;
}

static void
adns_hnl_preselect(HNLService *svc, SelectData *sd)
{
    AdnsHNLService *asvc = (AdnsHNLService *) svc;
	struct timeval *timeout;

	timeout = &sd->timeout;
	adns_beforeselect(asvc->adns, &sd->max_fd, &sd->read_fds,
			&sd->write_fds, &sd->except_fds, &timeout,
			NULL, NULL);
	sd->timeout.tv_usec = timeout->tv_usec;
	sd->timeout.tv_sec = timeout->tv_sec;
}

static void
adns_hnl_postselect(HNLService *svc, SelectData *sd)
{
    AdnsHNLService *asvc = (AdnsHNLService *) svc;

	adns_afterselect(asvc->adns, sd->max_fd, &sd->read_fds,
			&sd->write_fds, &sd->except_fds, NULL);

	if (asvc->lookup_count > 0) {
		AdnsHNLQueryData *data;
		adns_query query = NULL;
		adns_answer *answer = NULL;
		int rc;

		rc = adns_check(asvc->adns, &query, &answer, (void **) &data);
		/* EAGAIN is returned when no results are available yet. */
		/* ESRCH is returned if there are no queries. */

		if (rc == EAGAIN)
			return;
		if (rc != 0)
			die("cannot check DNS query: %s", strerror(rc));

		if (!data->cancelled) {
			if (answer->status == adns_s_ok)
				data->query->answer = xstrdup(*answer->rrs.str);
		    hnl_lookup_finished(svc->hmp, data->query);
		}

		asvc->lookup_count--;
		free(answer);
		free(data);
	}
}

HNLService *
adns_hnl_service_new(char *args)
{
    AdnsHNLService *asvc;

	if (args != NULL)
		return NULL;

	asvc = xmalloc(sizeof(AdnsHNLService));
    asvc->svc.initialize = NULL;
    asvc->svc.free = adns_hnl_free;
    asvc->svc.lookup = adns_hnl_lookup;
    asvc->svc.cancel = adns_hnl_cancel;
    asvc->svc.preselect = adns_hnl_preselect;
    asvc->svc.postselect = adns_hnl_postselect;
	asvc->lookup_count = 0;

	if (adns_init(&asvc->adns, adns_if_noerrprint, NULL) != 0)
		die("cannot initialize DNS facilities");

    return &asvc->svc;
}
