#include <config.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include "common/io-utils.h"
#include "common/error.h"
#include "common/memory.h"
#include "common/common.h"
#include "common/llist.h"
#include "libhnl-internal.h"

#define DEFAULT_PROCESS_COUNT 6

typedef struct {
	pid_t pid;
	int command_fd;
	int result_fd;
	HNLQuery *working_query;
	bool cancelled;
} GhbaHNLChild;

typedef struct {
	HNLService svc;
	int child_count;
	LList *pending_queries;
	LList *free_children;
	GhbaHNLChild child[0];
} GhbaHNLService;

static void
ghba_hnl_free(HNLService *svc)
{
	GhbaHNLService *gsvc = (GhbaHNLService *) svc;
	int c;

	for (c = 0; c < gsvc->child_count; c++) {
		close(gsvc->child[c].command_fd);
		close(gsvc->child[c].result_fd);
		waitpid(gsvc->child[c].pid, NULL, 0);
	}

	llist_free(gsvc->free_children);
	llist_free(gsvc->pending_queries);
	free(svc);
}

static void
ghba_hnl_write_child_command(GhbaHNLChild *child)
{
	uint32_t ip = child->working_query->ip;

	/* Assumes short write is error */
	if (write(child->command_fd, &ip, sizeof(ip)) < 0)
		die_errno("cannot write to child process");
}

static void
ghba_hnl_lookup(HNLService *svc, HNLQuery *query)
{
	GhbaHNLService *gsvc = (GhbaHNLService *) svc;

	if (!llist_is_empty(gsvc->free_children)) {
		GhbaHNLChild *child = llist_remove_first(gsvc->free_children);
		child->working_query = query;
		ghba_hnl_write_child_command(child);
	} else {
		llist_add(gsvc->pending_queries, query);
	}
}

static void
ghba_hnl_cancel(HNLService *svc, HNLQuery *query)
{
	GhbaHNLService *gsvc = (GhbaHNLService *) svc;
	int c;

	for (c = 0; c < gsvc->child_count; c++) {
		GhbaHNLChild *child = &gsvc->child[c];
		if (child->working_query == query) {
			child->cancelled = TRUE;
			return;
		}
	}

	llist_remove(gsvc->pending_queries, query);
}

static void
ghba_hnl_preselect(HNLService *svc, SelectData *sd)
{
	GhbaHNLService *gsvc = (GhbaHNLService *) svc;
	int c;

	for (c = 0; c < gsvc->child_count; c++) {
		if (gsvc->child[c].working_query != NULL) {
			FD_SET(gsvc->child[c].result_fd, &sd->read_fds);
	//		sd->max_fd = max(sd->max_fd, gsvc->child[c].result_fd+1);
			if (gsvc->child[c].result_fd >= sd->max_fd)
				sd->max_fd = gsvc->child[c].result_fd+1;
		}
	}
}

static void
ghba_hnl_postselect(HNLService *svc, SelectData *data)
{
	GhbaHNLService *gsvc = (GhbaHNLService *) svc;
	int c;

	for (c = 0; c < gsvc->child_count; c++) {
		GhbaHNLChild *child = &gsvc->child[c];

		if (FD_ISSET(child->result_fd, &data->read_fds)) {
			char result[65];

			/* Assumes short read is error */
			result[64] = '\0';
			if (read(child->result_fd, &result, sizeof(result)) < 0)
				die_errno("cannot read from child process");

			if (!child->cancelled) {
				child->working_query->answer = xstrdup(result);
				hnl_lookup_finished(svc->hmp, child->working_query);
			}

			child->cancelled = FALSE;
			if (llist_is_empty(gsvc->pending_queries)) {
				child->working_query = NULL;
				llist_append(gsvc->free_children, child);
			} else {
				child->working_query = llist_shift(gsvc->pending_queries);
				ghba_hnl_write_child_command(child);
			}
		}
	}
}

static void
ghba_child_main(int command_pipe, int result_pipe)
{
	while (TRUE) {
		char result[64];
		struct hostent *he;
		struct in_addr addr;

		/* Assumes short read is error */
		if (read(command_pipe, &addr.s_addr, sizeof(uint32_t)) < 0)
			exit(1);

		addr.s_addr = htonl(addr.s_addr);
		he = gethostbyaddr(&addr, sizeof(struct in_addr), AF_INET);
		if (he == NULL)
			result[0] = '\0';
		else
			strncpy(result, he->h_name, sizeof(result));

		/* Assumes short write is error */
		if (write(result_pipe, result, sizeof(result)) < 0)
			exit(1);
	}
}

HNLService *
ghba_hnl_service_new(char *args)
{
	GhbaHNLService *gsvc;
	int c;

	if (args == NULL)
		return NULL;

	gsvc = xmalloc(sizeof(GhbaHNLService)
					+ sizeof(GhbaHNLChild) * DEFAULT_PROCESS_COUNT);
	gsvc->svc.initialize = NULL;
	gsvc->svc.free = ghba_hnl_free;
	gsvc->svc.lookup = ghba_hnl_lookup;
	gsvc->svc.cancel = ghba_hnl_cancel;
	gsvc->svc.preselect = ghba_hnl_preselect;
	gsvc->svc.postselect = ghba_hnl_postselect;
	gsvc->child_count = DEFAULT_PROCESS_COUNT;
	gsvc->pending_queries = llist_new();
	gsvc->free_children = llist_new();

	for (c = 0; c < gsvc->child_count; c++) {
		pid_t child;
		int pipe_command[2];
		int pipe_result[2];

		if (pipe(pipe_command) < 0)
			die_errno("cannot create communication pipe");
		if (pipe(pipe_result) < 0)
			die_errno("cannot create communication pipe");

		child = fork();
		if (child == -1)
			die_errno("cannot create child process");
		if (child == 0) {
			close(pipe_command[1]);
			close(pipe_result[0]);
			ghba_child_main(pipe_command[0], pipe_result[1]);
		}

		close(pipe_command[0]);
		close(pipe_result[1]);
		gsvc->child[c].pid = child;
		gsvc->child[c].command_fd = pipe_command[1];
		gsvc->child[c].result_fd = pipe_result[0];
		gsvc->child[c].working_query = NULL;
		gsvc->child[c].cancelled = FALSE;
		llist_add(gsvc->free_children, &gsvc->child[c]);
	}

	return &gsvc->svc;
}
