#include <config.h>
#include <stdio.h>
#include <ctype.h>
#include "libhnl-internal.h"
#include "common/error.h"
#include "common/memory.h"
#include "common/hmap.h"

#define DEFAULT_HOSTS_FILE "/etc/hosts"

typedef struct {
    HNLService svc;
    HMap *hosts_map;
} HostsHNLService;

static uint32_t
hosts_map_hash_key(const void *key)
{
	return (uint32_t) key;
}

static int32_t
hosts_map_compare_keys(const void *key1, const void *key2)
{
	return (uint32_t) key1 - (uint32_t) key2;
}

static void
hosts_hnl_lookup(HNLService *svc, HNLQuery *query)
{
    HostsHNLService *hsvc = (HostsHNLService *) svc;
    char *name;

	name = hmap_get(hsvc->hosts_map, (void *) query->ip);
	if (name != NULL)
		query->answer = xstrdup(name);

	hnl_lookup_finished(svc->hmp, query);
}

static void
hosts_hnl_free(HNLService *svc)
{
    HostsHNLService *hsvc = (HostsHNLService *) svc;

	hmap_iterate_values(hsvc->hosts_map, (IteratorFunc) free);
    hmap_free(hsvc->hosts_map);
    free(hsvc);
}

static void
hosts_hnl_service_read_hosts(HNLService *svc, const char *filename)
{
    HostsHNLService *hsvc = (HostsHNLService *) svc;
	FILE *file;
	char *line = NULL;
	size_t linesize = 0;

	file = fopen(filename, "r");
	if (file == NULL)
		die_errno("%s: cannot open file", filename);

	while (getline(&line, &linesize, file) > 0) {
		uint32_t ip;
		int a, b, c, d;
		char *ipstr;
		char *name;
		char *t;

		for (t = line; isblank(*t); t++);
		if (*t == '\0' || *t == '#' || *t == '\n')
			continue;
		ipstr = t;

		for (; !isblank(*t) && *t != '#'; t++);
		if (*t == '\0' || *t == '#' || *t == '\n')
			continue;
		*t = '\0';

		for (t++; isblank(*t); t++);
		if (*t == '\0' || *t == '#' || *t == '\n')
			continue;
		name = t;

		for (; !isblank(*t) && *t != '#' && *t != '\n'; t++);
		*t = '\0';

		if (sscanf(ipstr, "%d.%d.%d.%d", &a, &b, &c, &d) != 4)
			continue;
		
		ip = ((a & 0xFF)<<24) | ((b & 0xFF)<<16) | ((c & 0xFF)<<8) | (d & 0xFF);
		hmap_put(hsvc->hosts_map, (void *) ip, xstrdup(name));
	}

	if (line != NULL)
		free(line);
	if (ferror(file))
		die_errno("%s: cannot read file", filename);
	fclose(file);
}

HNLService *
hosts_hnl_service_new(char *args)
{
    HostsHNLService *hsvc = xmalloc(sizeof(HostsHNLService));

    hsvc->svc.initialize = NULL;
    hsvc->svc.free = hosts_hnl_free;
    hsvc->svc.lookup = hosts_hnl_lookup;
    hsvc->svc.cancel = NULL;
    hsvc->svc.preselect = NULL;
    hsvc->svc.postselect = NULL;

    hsvc->hosts_map = hmap_new();
	hmap_set_hash_function(hsvc->hosts_map, hosts_map_hash_key);
	hmap_set_compare_function(hsvc->hosts_map, hosts_map_compare_keys);

	if (args == NULL) {
		hosts_hnl_service_read_hosts(&hsvc->svc, DEFAULT_HOSTS_FILE);
	} else if (strcmp(args, "") != 0) {
		hosts_hnl_service_read_hosts(&hsvc->svc, args);
	}

    return &hsvc->svc;
}
