#include "common/common.h"
#include "common/memory.h"
#include "common/hmap.h"
#include "common/error.h"
#include "common/string-utils.h"
#include "libhnl-services.h"
#include "libhnl-internal.h"

static bool query_finished;

static uint32_t
cache_map_hash_key(const void *key)
{
	return (uint32_t) key;
}

static int32_t
cache_map_compare_keys(const void *key1, const void *key2)
{
	return (uint32_t) key1 - (uint32_t) key2;
}

HNLManager *
hnl_new(void)
{
    HNLManagerPrivate *hmp = xmalloc(sizeof(HNLManagerPrivate));

    hmp->services = NULL;
    hmp->service_count = 0;

	hmp->caching = FALSE;
	hmp->cache = hmap_new();
	hmap_set_hash_function(hmp->cache, cache_map_hash_key);
	hmap_set_compare_function(hmp->cache, cache_map_compare_keys);

    return &hmp->hm;
}

bool
hnl_initialize_from_spec(HNLManager *hm, const char *spec)
{
	while (TRUE) {
		HNLService *svc;
		const char *next = strchr(spec, ',');
		char *args = strchr(spec, ':');
		int len;

		if (next == NULL)
			next = spec + strlen(spec);

		if (args != NULL && args < next) {
			len = args-spec;
			args = substring(spec, args-spec+1, next-spec);
		} else {
			len = next-spec;
			args = NULL;
		}

		if (strncmp(spec, "hosts", len) == 0) {
			svc = hosts_hnl_service_new(args);
		} else if (strncmp(spec, "adns", len) == 0) {
			svc = adns_hnl_service_new(args);
		} else if (strncmp(spec, "ghba", len) == 0) {
			svc = ghba_hnl_service_new(args);
		} else {
			set_error("no such hostname service");
			svc = NULL;
		}

		if (args != NULL)
			free(args);
		if (svc == NULL)
			return FALSE;

		hnl_add_service(hm, svc);
		if (*next == '\0')
			return TRUE;
		spec = next+1;
	}
}

void
hnl_set_caching(HNLManager *hm, bool enable)
{
    HNLManagerPrivate *hmp = (HNLManagerPrivate *) hm;
	hmp->caching = enable;
}

void
hnl_free(HNLManager *hm)
{
	HNLManagerPrivate *hmp = (HNLManagerPrivate *) hm;
	int c;

	for (c = 0; c < hmp->service_count; c++) {
		HNLService *svc = hmp->services[c];
		if (svc->free != NULL)
			svc->free(svc);
	}

	if (hmp->services != NULL)
		free(hmp->services);

	free(hmp);
}

void
hnl_add_service(HNLManager *hm, HNLService *svc)
{
    HNLManagerPrivate *hmp = (HNLManagerPrivate *) hm;

    hmp->service_count++;
    hmp->services = xrealloc(hmp->services, hmp->service_count * sizeof(void *));
    hmp->services[hmp->service_count-1] = svc;
	svc->hmp = hmp;
    if (svc->initialize != NULL)
    	svc->initialize(svc);
}

HNLQuery *
hnl_lookup(HNLManager *hm, uint32_t ip, void *userdata)
{
    HNLManagerPrivate *hmp = (HNLManagerPrivate *) hm;
    HNLQuery *query;

	if (hmp->caching && hmap_contains_key(hmp->cache, (void *) ip)) {
		char *answer = hmap_get(hmp->cache, (void *) ip);
		hmp->hm.lookup_finished(&hmp->hm, ip, answer, userdata);
		return NULL;
	}

    query = xmalloc(sizeof(HNLQuery));
    query->ip = ip;
    query->answer = NULL;
    query->userdata = userdata;
    query->current_service = -1;
	query_finished = FALSE;

    hnl_lookup_finished(hmp, query);

	return (query_finished ? NULL : query);
}

void
hnl_lookup_finished(HNLManagerPrivate *hmp, HNLQuery *query)
{
    HNLService *svc;

    query->current_service++;
    if (query->answer != NULL || query->current_service >= hmp->service_count) {
		query_finished = TRUE;
		hmp->hm.lookup_finished(&hmp->hm, query->ip, query->answer, query->userdata);
		if (hmp->caching) {
			hmap_put(hmp->cache, (void *) query->ip, query->answer);
		} else if (query->answer != NULL) {
	    	free(query->answer);
		}
		free(query);
    } else {
	    svc = hmp->services[query->current_service];
	    svc->lookup(svc, query);
	}
}

void
hnl_cancel(HNLManager *hm, HNLQuery *query)
{
    HNLManagerPrivate *hmp = (HNLManagerPrivate *) hm;
    HNLService *svc = hmp->services[query->current_service];

    if (svc->cancel != NULL)
    	svc->cancel(svc, query);
	free(query);
}

void
hnl_preselect(HNLManager *hm, SelectData *data)
{
    HNLManagerPrivate *hmp = (HNLManagerPrivate *) hm;
    int c;

    for (c = 0; c < hmp->service_count; c++) {
    	HNLService *svc = hmp->services[c];
		if (svc->preselect != NULL)
	    	svc->preselect(svc, data);
    }
}

void
hnl_postselect(HNLManager *hm, SelectData *data)
{
    HNLManagerPrivate *hmp = (HNLManagerPrivate *) hm;
    int c;

    for (c = 0; c < hmp->service_count; c++) {
    	HNLService *svc = hmp->services[c];
		if (svc->postselect != NULL)
	    	svc->postselect(svc, data);
    }
}

void
hnl_clear_select_data(SelectData *sd)
{
	FD_ZERO(&sd->read_fds);
	FD_ZERO(&sd->write_fds);
	FD_ZERO(&sd->except_fds);
	sd->max_fd = 0;
	sd->timeout.tv_sec = 0;
	sd->timeout.tv_usec = 0;
}
