#include <config.h>
#include <stdio.h>
#include "common/string-utils.h"
#include "common/memory.h"
#include "common/error.h"
#include "common/intparse.h"
#include "libpnl.h"

struct _PNLManager {
	char *tcp_port_names[UINT16_MAX];
	char *udp_port_names[UINT16_MAX];
};

PNLManager *
pnl_new(void)
{
	PNLManager *pm = xmalloc(sizeof(PNLManager));
	memset(pm, 0, sizeof(PNLManager));
	return pm;
}

static void
pnl_load_port_spec(PNLManager *pm, const char *filename)
{
	FILE *file;
	char *line = NULL;
	size_t linesize = 0;

	file = fopen(filename, "r");
	if (file == NULL)
		die_errno("%s: cannot open file", filename);

	while (getline(&line, &linesize, file) > 0) {
		char *name;
		char *port;
		char *proto;
		char *t;
		uint16_t portnum;

		for (t = line; isblank(*t); t++);
		if (*t == '\0' || *t == '#' || *t == '\n')
			continue;

		name = t;
		for (; !isblank(*t) && *t != '#'; t++);
		if (*t == '\0' || *t == '#' || *t == '\n')
			continue;
		*t = '\0';

		for (t++; isblank(*t); t++);
		if (*t == '\0' || *t == '#' || *t == '\n')
			continue;
		port = t;
		for (; !isblank(*t) && *t != '#' && *t != '\n' && *t != '/'; t++);
		
		if (*t == '/') {
			*t = '\0';
			proto = ++t;
			for (; !isblank(*t) && *t != '#' && *t != '\n'; t++);
			*t = '\0';
		} else {
			*t = '\0';
			proto = NULL;
		}

		if (!parse_uint16(port, &portnum))
			continue;

		if (proto == NULL || strcmp(proto, "tcp") == 0)
			pm->tcp_port_names[portnum] = xstrdup(name);
		if (proto == NULL || strcmp(proto, "udp") == 0)
			pm->udp_port_names[portnum] = xstrdup(name);
	}

	if (line != NULL)
		free(line);
	if (ferror(file))
		die_errno("%s: cannot read file", filename);
	fclose(file);
}

bool
pnl_initialize_from_spec(PNLManager *pm, const char *spec)
{
	if (*spec == '\0')
		return TRUE;

	while (TRUE) {
		const char *next = strchr(spec, ':');
		char *file;

		if (next == NULL)
			next = spec + strlen(spec);
		if (next == spec) {
			set_error("invalid port file specification");
			return FALSE;
		}

		file = substring(spec, 0, next-spec);
		pnl_load_port_spec(pm, file);
		free(file);

		if (*next == '\0')
			return TRUE;
		spec = next+1;
	}

	return TRUE;
}

const char *
pnl_lookup(PNLManager *pm, uint16_t protocol, uint16_t port)
{
	if (protocol == PNL_PROTO_TCP)
		return pm->tcp_port_names[port];
	if (protocol == PNL_PROTO_UDP)
		return pm->tcp_port_names[port];
	return NULL;
}

void
pnl_free(PNLManager *pm)
{
	int c;

	for (c = 0; c < UINT16_MAX; c++) {
		if (pm->tcp_port_names[c] != NULL)
			free(pm->tcp_port_names[c]);
		if (pm->udp_port_names[c] != NULL)
			free(pm->udp_port_names[c]);
	}

	free(pm);
}
